
// *****************************************
//   DELUGE - Back-End Full Search
//
//   ver:1.1
//   status: 1-memory case under test   Mar/07
//   notes: (only 1-mem unit uses npid)
//
// *****************************************
//   Verilog Code (30 Mar 07)
//   Copyright Iain Devlin 2006-07
//
// *****************************************




// *****************************************
//   Deluge - Back-End Search Unit
//
// *****************************************

module Del_BackEndFullSearch (pid,id,vKey,rKey,IV,datain,adr,wr,em, rd,oKey,keyflag, clk,pause,aclr, xclk);
parameter  i = 2, j = 4, v = 31, h = 0,
           w = 16, k = 80;
`ifdef DEL_NO_ID parameter nid = 1, xnid = 0; `else parameter nid = 0, xnid = 1; `endif
`ifdef DEL_NO_PID parameter npid = 1, xnpid = 0; `else parameter npid = 0, xnpid = 1; `endif

// =========================================

 input [j+npid-h-1:0] pid;
 input [i+nid-1:0] id;
 input [v-1:0] vKey;
 input [k-v-i-1:0] rKey;
 input [32-1:0] IV;

 input [w-1:0] datain;
 input [j+2:0] adr;
 input  wr;

 input  em;
 output rd;

 output [k-1:0] oKey;
 output keyflag;

 input xclk, clk,pause,aclr;

 wire rekey;

 wire [w-1:0] keystream;
 wire valid;
 reg  [w-1:0] keystream_o;
 reg  valid_o;

 wire [2:0] ra;
 wire [j+2:0] radr;
 wire [w-1:0] plain;

 wire cmp,sync;
 wire next;
 
 // - - - - - - - - - - - - - - - - - - - - - -
 // for system debug add
 // - - - - - - - - - - - - - - - - - - - - - -
 // * module head: ,plainstream_);
 // - - - - - - - - - - - - - - - - - - - - - -
 // * output [w-1:0] plainstream_;
 // - - - - - - - - - - - - - - - - - - - - - -
 // + assign plainstream_ = plain0;
 // - - - - - - - - - - - - - - - - - - - - - -

// =========================================

 Del_RekeyCTRL RCTL (next,em, rekey,rd, clk,pause,aclr);
 `DEL_STREAM_CIPHER_B #(.w(w))  SC0 ({{(xnid){id}},vKey,rKey},IV,rekey,pause, keystream,valid, clk,aclr);

 Del_Count RA0 (valid, ra, clk,pause,aclr); 
 assign radr = {{(xnpid){pid}},ra};

 Del_FullMemD #(.j(j),.w(w)) MEM0 (datain,adr,wr,radr, plain, clk,pause,aclr, xclk);

 always @(posedge clk, posedge aclr) begin
   if(aclr)        keystream_o <= 'b0;
   else if(!pause) keystream_o <= keystream;
  end
 always @(posedge clk, posedge aclr) begin
   if(aclr)        valid_o <= 'b0;
   else if(!pause) valid_o <= valid;
  end

 Del_CMP_s2b #(.w(w)) CMP0 (plain,keystream_o,valid_o, cmp,sync, clk,pause,aclr);

 Del_FlagCTRL FCTL (cmp,sync, keyflag,next, clk,pause,aclr);

 assign oKey = {{(xnid){id}},vKey,rKey};

// =========================================

endmodule





// ------------------------------------------------------
// ------------------------------------------------------







// *****************************************
//   Deluge - Rekey Control
//
// *****************************************

module Del_RekeyCTRL (next,empty, rekey,read, clk,pause,aclr);


// =========================================

 input  next,empty;
 output reg rekey,read;

 input clk,pause,aclr;

 reg  [2:0] state;

// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr) begin
       state <= 3'b0;
       read  <= 1'b0;
       rekey <= 1'b1;
      end
   else if(!pause) 
     casex({state,next,empty})
       5'b000x0 :  begin state <= 3'b001; read <= 1'b1; rekey <= 1'b1; end
       5'b000x1 :  begin state <= 3'b000; read <= 1'b0; rekey <= 1'b1; end
       5'b001xx :  begin state <= 3'b101; read <= 1'b0; rekey <= 1'b1; end
       5'b1010x :  begin state <= 3'b011; read <= 1'b0; rekey <= 1'b0; end
       5'b1011x :  begin state <= 3'b101; read <= 1'b0; rekey <= 1'b0; end
       5'b0110x :  begin state <= 3'b011; read <= 1'b0; rekey <= 1'b0; end
       5'b01110 :  begin state <= 3'b001; read <= 1'b1; rekey <= 1'b0; end
       5'b01111 :  begin state <= 3'b010; read <= 1'b0; rekey <= 1'b1; end
       5'b010x0 :  begin state <= 3'b001; read <= 1'b1; rekey <= 1'b0; end
       5'b010x1 :  begin state <= 3'b010; read <= 1'b0; rekey <= 1'b1; end
     default begin state <= 3'b00; read <= 1'b0; rekey <= 1'b1; end endcase
  end  


// =========================================

endmodule





// *****************************************
//   Deluge - Address Count
//
// *****************************************

module Del_Count (en, count, clk,pause,aclr);
parameter k = 3, x = 5;

// =========================================

 input  en;
 output reg [k-1:0] count;

 input clk,pause,aclr;

// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)           count <= 'b0;
   else if(!en)       count <= 'b0;
   else if(!pause)
     if(count==(x-1)) count <= 'b0;
     else             count <= count + 1'b1;
  end

// =========================================

endmodule





// *****************************************
//   Deluge -  Dual Clocked Memory
//
// *****************************************

module Del_FullMemD (d,wa,we,ra, q, clk,pause,aclr, xclk);
parameter j = 4, p = 2**j, w = 16, k = 3;
          
// =========================================

 input [w-1:0] d;
 input [j+2:0] wa,ra;
 input we;

 output reg [w-1:0] q;

 input xclk, clk,pause,aclr;

 reg [w-1:0] mem [((2**k)*p)-1:0];
 //(* ram_init_file = "DEL_MEM.mif" *) reg [w-1:0] mem [((2**k)*p)-1:0];

// =========================================
  
 always @(posedge xclk)  begin
   if(we) mem[wa] <= d;
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        q <= 'b0;
   else if(!pause) q <= mem[ra];
  end

// =========================================

endmodule






// *****************************************
//   Compare - 2 stage, short front stage
//
// *****************************************

module Del_CMP_s2b (keystream,plainstream,k_valid, match,m_valid, clk,pause,aclr);
parameter w = 16,  //data width in compares
          c = 8,   //num of front cmp stages
          x = w/c; //front stage width

// =========================================

 input  [w-1:0]  keystream,plainstream;
 input  k_valid;

 output reg match;
 output reg m_valid;

 input clk,pause,aclr;

 reg  [c-1:0] cmp;
 reg  c_valid;

// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)        cmp <= 'b0;
   else if(!pause)
     if(!k_valid)  cmp <= 'b0;
     else begin
       cmp[7] <= &( keystream[7*x+x-1:7*x] ~^ plainstream[7*x+x-1:7*x] );
       cmp[6] <= &( keystream[6*x+x-1:6*x] ~^ plainstream[6*x+x-1:6*x] );
       cmp[5] <= &( keystream[5*x+x-1:5*x] ~^ plainstream[5*x+x-1:5*x] );
       cmp[4] <= &( keystream[4*x+x-1:4*x] ~^ plainstream[4*x+x-1:4*x] );
       cmp[3] <= &( keystream[3*x+x-1:3*x] ~^ plainstream[3*x+x-1:3*x] );
       cmp[2] <= &( keystream[2*x+x-1:2*x] ~^ plainstream[2*x+x-1:2*x] );
       cmp[1] <= &( keystream[1*x+x-1:1*x] ~^ plainstream[1*x+x-1:1*x] );
       cmp[0] <= &( keystream[0*x+x-1:0*x] ~^ plainstream[0*x+x-1:0*x] );
     end
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        c_valid <= 'b0;
   else if(!pause) c_valid <= k_valid;
  end


 always @(posedge clk, posedge aclr) begin
   if(aclr)         match <= 'b0;
   else if(!pause)   
     if(!c_valid)   match <= 'b0;
     else           match <= &cmp;  
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        m_valid <= 'b0;
   else if(!pause) m_valid <= c_valid;
  end


// =========================================

endmodule






// *****************************************
//   Deluge - Keyflag Control
//
// *****************************************

module Del_FlagCTRL (cmp,sync, flag,next, clk,pause,aclr);
parameter k = 3, x = 5;

// =========================================

 input  cmp,sync;
 output reg flag,next;

 input clk,pause,aclr;

 reg [k-1:0] count;

// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)           count <= 'b0;
   else if(!sync)     count <= 'b0;
   else if(!pause)
      if(count!=(x))  count <= count + cmp;
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)           flag <= 'b0;
   else if(!sync)     flag <= 'b0; 
   else if(!pause)
      if(count==(x)&&(!next)) flag <= 1'b1;
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)           next <= 'b0;
   else if(!sync)     next <= 'b0;
   else if(!pause)
     if(!cmp || flag) next <= 1'b1;
  end

// =========================================

endmodule





// ------------------------------------------------------
// ------------------------------------------------------








// *****************************************
//   Back-End Search Module   2x memory
//   (Fully Tested)
// *****************************************

module Del_BackEndFullSearch2 (pid,id,vKey,rKey,IV,datain,adr,wr,em, rd,oKey,keyflag, clk,pause,aclr, xclk);
parameter  i = 2, j = 4, v = 31, h = 1,
           w = 16, k = 80, s = 2**h;
`ifdef DEL_NO_ID parameter nid = 1, xnid = 0; `else parameter nid = 0, xnid = 1; `endif

// =========================================

 input [j-h-1:0] pid;
 input [i+nid-1:0] id;
 input [v-1:0] vKey;
 input [k-v-i-1:0] rKey;
 input [32-1:0] IV;

 input [w-1:0] datain;
 input [j+2:0] adr;
 input  wr;

 input  em;
 output rd;

 output [k-1:0] oKey;
 output keyflag;

 input xclk, clk,pause,aclr;

 wire  rekey;

 wire [w-1:0] keystream;
 wire valid;
 reg  [w-1:0] keystream_o;
 reg  valid_o;

 wire [2:0] ra;
 wire [j+2-h:0] radr;
 wire [w-1:0] plain1,plain0;

 wire [s-1:0] cmp;
 wire sync;
 wire next;
 
 // - - - - - - - - - - - - - - - - - - - - - -
 // for system debug add
 // - - - - - - - - - - - - - - - - - - - - - -
 // * module head: ,plainstream_,plainstream1_);
 // - - - - - - - - - - - - - - - - - - - - - -
 // * output [w-1:0] plainstream_,plainstream1_;
 // - - - - - - - - - - - - - - - - - - - - - -
 // * assign plainstream_  = plain0;
 // * assign plainstream1_ = plain1;
 // - - - - - - - - - - - - - - - - - - - - - -

// =========================================

 Del_RekeyCTRL RCTL (next,em, rekey,rd, clk,pause,aclr);
 `DEL_STREAM_CIPHER_B #(.w(w))  SC1 ({{(xnid){id}},vKey,rKey},IV,rekey,pause, keystream,valid, clk,aclr);

 Del_Count RA0 (valid, ra, clk,pause,aclr); 
 assign radr = {pid,ra};

 Del_FullMemD #(.j(j-h),.w(w)) MEM1 (datain,{adr[j+2:4],adr[2:0]},{&{wr,  adr[3] }},radr, plain1, clk,pause,aclr, xclk);
 Del_FullMemD #(.j(j-h),.w(w)) MEM0 (datain,{adr[j+2:4],adr[2:0]},{&{wr,{~adr[3]}}},radr, plain0, clk,pause,aclr, xclk);
 //Del_FullMemoryD  MEM1 (datain,aclr,radr,clk, aclr,{adr[j+2:4],adr[2:0]},xclk,{&{wr,  adr[3] }}, plain1); //avoid: as has added clk
 //Del_FullMemoryD  MEM0 (datain,aclr,radr,clk, aclr,{adr[j+2:4],adr[2:0]},xclk,{&{wr,{~adr[3]}}}, plain0); //delay vs designed system

 always @(posedge clk, posedge aclr) begin
   if(aclr)        keystream_o <= 'b0;
   else if(!pause) keystream_o <= keystream;
  end
 always @(posedge clk, posedge aclr) begin
   if(aclr)        valid_o <= 'b0;
   else if(!pause) valid_o <= valid;
  end

 Del_CMP_s2b #(.w(w)) CMP1 (plain1,keystream_o,valid_o, cmp[1],    , clk,pause,aclr);
 Del_CMP_s2b #(.w(w)) CMP0 (plain0,keystream_o,valid_o, cmp[0],sync, clk,pause,aclr);

 Del_FlagCTRL FCTL ({|cmp},sync, keyflag,next, clk,pause,aclr);

 assign oKey = {{(xnid){id}},vKey,rKey};

// =========================================

endmodule






// *****************************************
//   Back-End Search Module   4x memory
//
// *****************************************

module Del_BackEndFullSearch4 (pid,id,vKey,rKey,IV,datain,adr,wr,em, rd,oKey,keyflag, clk,pause,aclr, xclk);
parameter  i = 2, j = 4, v = 31, h = 2,
           w = 16, k = 80, s = 2**h;
`ifdef DEL_NO_ID parameter nid = 1, xnid = 0; `else parameter nid = 0, xnid = 1; `endif

// =========================================

 input [j-h-1:0] pid;
 input [i+nid-1:0] id;

 input [v-1:0] vKey;
 input [k-v-i-1:0] rKey;
 input [32-1:0] IV;

 input [w-1:0] datain;
 input [j+2:0] adr;
 input  wr;

 input  em;
 output rd;

 output [k-1:0] oKey;
 output keyflag;

 input xclk, clk,pause,aclr;

 wire rekey;

 wire [w-1:0] keystream;
 wire valid;
 reg  [w-1:0] keystream_o;
 reg  valid_o;

 wire [2:0] ra;
 wire [j+2-h:0] radr;
 wire [w-1:0] plain3,plain2,plain1,plain0;

 wire [s-1:0] cmp;
 wire sync;
 wire next;
 
 // - - - - - - - - - - - - - - - - - - - - - -
 // for system debug add
 // - - - - - - - - - - - - - - - - - - - - - -
 // * module head: ,plainstream_,plainstream1_);
 // - - - - - - - - - - - - - - - - - - - - - -
 // * output [w-1:0] plainstream_,plainstream1_;
 // - - - - - - - - - - - - - - - - - - - - - -
 // + assign plainstream_  = plain0;
 // + assign plainstream1_ = plain1;
 // - - - - - - - - - - - - - - - - - - - - - -

// =========================================

 Del_RekeyCTRL RCTL (next,em, rekey,rd, clk,pause,aclr);
 `DEL_STREAM_CIPHER_B #(.w(w))  SC1 ({{(xnid){id}},vKey,rKey},IV,rekey,pause, keystream,valid, clk,aclr);

 Del_Count RA0 (valid, ra, clk,pause,aclr); 
 assign radr = {pid,ra};

 Del_FullMemD #(.j(j-h),.w(w)) MEM3 (datain,{adr[j+2:5],adr[2:0]},{ &{wr,  adr[4] ,  adr[3] } },radr, plain3, clk,pause,aclr, xclk);
 Del_FullMemD #(.j(j-h),.w(w)) MEM2 (datain,{adr[j+2:5],adr[2:0]},{ &{wr,  adr[4] ,{~adr[3]}} },radr, plain2, clk,pause,aclr, xclk);
 Del_FullMemD #(.j(j-h),.w(w)) MEM1 (datain,{adr[j+2:5],adr[2:0]},{ &{wr,{~adr[4]},  adr[3] } },radr, plain1, clk,pause,aclr, xclk);
 Del_FullMemD #(.j(j-h),.w(w)) MEM0 (datain,{adr[j+2:5],adr[2:0]},{ &{wr,{~adr[4]},{~adr[3]}} },radr, plain0, clk,pause,aclr, xclk);

 always @(posedge clk, posedge aclr) begin
   if(aclr)        keystream_o <= 'b0;
   else if(!pause) keystream_o <= keystream;
  end
 always @(posedge clk, posedge aclr) begin
   if(aclr)        valid_o <= 'b0;
   else if(!pause) valid_o <= valid;
  end

 Del_CMP_s2b #(.w(w)) CMP3 (plain3,keystream_o,valid_o, cmp[3],    , clk,pause,aclr);
 Del_CMP_s2b #(.w(w)) CMP2 (plain2,keystream_o,valid_o, cmp[2],    , clk,pause,aclr);
 Del_CMP_s2b #(.w(w)) CMP1 (plain1,keystream_o,valid_o, cmp[1],    , clk,pause,aclr);
 Del_CMP_s2b #(.w(w)) CMP0 (plain0,keystream_o,valid_o, cmp[0],sync, clk,pause,aclr);


 Del_FlagCTRL FCTL ({|cmp},sync, keyflag,next, clk,pause,aclr);

 assign oKey = {{(xnid){id}},vKey,rKey};

// =========================================

endmodule






// *****************************************
//   Back-End Search Module   8x memory
//
// *****************************************

module Del_BackEndFullSearch8 (pid,id,vKey,rKey,IV,datain,adr,wr,em, rd,oKey,keyflag, clk,pause,aclr, xclk);
parameter  i = 2, j = 4, v = 31, h = 3,
           w = 16, k = 80, s = 2**h;
`ifdef DEL_NO_ID parameter nid = 1, xnid = 0; `else parameter nid = 0, xnid = 1; `endif

// =========================================

 input [j-h-1:0] pid;
 input [i+nid-1:0] id;
 input [v-1:0] vKey;
 input [k-v-i-1:0] rKey;
 input [32-1:0] IV;

 input [w-1:0] datain;
 input [j+2:0] adr;
 input  wr;

 input  em;
 output rd;

 output [k-1:0] oKey;
 output keyflag;

 input xclk, clk,pause,aclr;

 wire rekey;

 wire [w-1:0] keystream;
 wire valid;
 reg  [w-1:0] keystream_o;
 reg  valid_o;

 wire [2:0] ra;
 wire [j+2-h:0] radr;
 wire [s-1:0] swr;
 wire [w-1:0] plain7,plain6,plain5,plain4,plain3,plain2,plain1,plain0;

 wire [s-1:0] cmp;
 wire sync;
 wire next;
 
// =========================================

 Del_RekeyCTRL RCTL (next,em, rekey,rd, clk,pause,aclr);
 `DEL_STREAM_CIPHER_B #(.w(w))  SC1 ({{(xnid){id}},vKey,rKey},IV,rekey,pause, keystream,valid, clk,aclr);

 Del_Count RA0 (valid, ra, clk,pause,aclr); 
 assign radr = {pid,ra};

 assign swr[0] = { &{wr,  adr[5],   adr[4],   adr[3] } };
 assign swr[1] = { &{wr,  adr[5],   adr[4], {~adr[3]}} };
 assign swr[2] = { &{wr,  adr[5], {~adr[4]},  adr[3] } };
 assign swr[3] = { &{wr,  adr[5], {~adr[4]},{~adr[3]}} };
 assign swr[4] = { &{wr,{~adr[5]},  adr[4],   adr[3] } };
 assign swr[5] = { &{wr,{~adr[5]},  adr[4], {~adr[3]}} };
 assign swr[6] = { &{wr,{~adr[5]},{~adr[4]},  adr[3] } };
 assign swr[7] = { &{wr,{~adr[5]},{~adr[4]},{~adr[3]}} };
 
 Del_FullMemD #(.j(j-h),.w(w)) MEM7 (datain,{adr[j+2:6],adr[2:0]},swr[0],radr, plain7, clk,pause,aclr, xclk);
 Del_FullMemD #(.j(j-h),.w(w)) MEM6 (datain,{adr[j+2:6],adr[2:0]},swr[1],radr, plain6, clk,pause,aclr, xclk);
 Del_FullMemD #(.j(j-h),.w(w)) MEM5 (datain,{adr[j+2:6],adr[2:0]},swr[2],radr, plain5, clk,pause,aclr, xclk);
 Del_FullMemD #(.j(j-h),.w(w)) MEM4 (datain,{adr[j+2:6],adr[2:0]},swr[3],radr, plain4, clk,pause,aclr, xclk);
 Del_FullMemD #(.j(j-h),.w(w)) MEM3 (datain,{adr[j+2:6],adr[2:0]},swr[4],radr, plain3, clk,pause,aclr, xclk);
 Del_FullMemD #(.j(j-h),.w(w)) MEM2 (datain,{adr[j+2:6],adr[2:0]},swr[5],radr, plain2, clk,pause,aclr, xclk);
 Del_FullMemD #(.j(j-h),.w(w)) MEM1 (datain,{adr[j+2:6],adr[2:0]},swr[6],radr, plain1, clk,pause,aclr, xclk);
 Del_FullMemD #(.j(j-h),.w(w)) MEM0 (datain,{adr[j+2:6],adr[2:0]},swr[7],radr, plain0, clk,pause,aclr, xclk);

 always @(posedge clk, posedge aclr) begin
   if(aclr)        keystream_o <= 'b0;
   else if(!pause) keystream_o <= keystream;
  end
 always @(posedge clk, posedge aclr) begin
   if(aclr)        valid_o <= 'b0;
   else if(!pause) valid_o <= valid;
  end

 Del_CMP_s2b #(.w(w)) CMP7 (plain7,keystream_o,valid_o, cmp[7],    , clk,pause,aclr);
 Del_CMP_s2b #(.w(w)) CMP6 (plain6,keystream_o,valid_o, cmp[6],    , clk,pause,aclr);
 Del_CMP_s2b #(.w(w)) CMP5 (plain5,keystream_o,valid_o, cmp[5],    , clk,pause,aclr);
 Del_CMP_s2b #(.w(w)) CMP4 (plain4,keystream_o,valid_o, cmp[4],    , clk,pause,aclr);
 Del_CMP_s2b #(.w(w)) CMP3 (plain3,keystream_o,valid_o, cmp[3],    , clk,pause,aclr);
 Del_CMP_s2b #(.w(w)) CMP2 (plain2,keystream_o,valid_o, cmp[2],    , clk,pause,aclr);
 Del_CMP_s2b #(.w(w)) CMP1 (plain1,keystream_o,valid_o, cmp[1],    , clk,pause,aclr);
 Del_CMP_s2b #(.w(w)) CMP0 (plain0,keystream_o,valid_o, cmp[0],sync, clk,pause,aclr);


 Del_FlagCTRL FCTL ({|cmp},sync, keyflag,next, clk,pause,aclr);

 assign oKey = {{(xnid){id}},vKey,rKey};

// =========================================

endmodule