
// *****************************************
//   DELUGE - Front-End Search
//
//   ver:1.1
//   status: 4x2 under test   Mar/07
//   notes: (adds a dual quad module)
//
// *****************************************
//   Verilog Code (30 Mar 07)
//   Copyright Iain Devlin 2006-07
//
// *****************************************


`define DEL_CMP Del_CMP_p 
//Del_CMP_p   //2 stage, fully variable 

`define DEL_WRCTRL Del_wrCTRL_s3c
//Del_wrCTRL_s3c //max 4096 flags but uses compaction on output 
//Del_wrCTRL_s3  //max 4096 flags - no longer in module
//Del_wrCTRL_s2c //max 256 flags but uses compaction on output 
//Del_wrCTRL_s2  //max 256 flags - no longer in module
//Del_wrCTRL_s1  //max 16 flags
//
`define DEL_WR_VKEY_DELAY `DEL_WR_VKEY_DELAYs3   //s1,s2,s3
`define DEL_WR_VKEY_DELAYs1 always@* begin vsync<=sync; end
`define DEL_WR_VKEY_DELAYs2 always@(posedge clk,posedge aclr) begin if(aclr) vsync<=1'b0; else if(!pause) vsync<=sync; end
`define DEL_WR_VKEY_DELAYs3 always@(posedge clk,posedge aclr) begin if(aclr)begin wsync<='b0; vsync<='b0;end else if(!pause)begin wsync<=sync;vsync<=wsync;end end



// *****************************************
//   Deluge - Front-End Search Unit
//
// *****************************************

module Del_FrontEndSearch (rKey,IV,set,plainMem, fd,vd,done,em, read, clk,pause,aclr);
parameter  i = 0, j = 4, v = 31, h = 2, w = 16, k = 80,
           n = 2**i,  //number of ciphers
           p = 2**j,  //number of plainstreams
           s = 2**h,  //compaction factor
           f = p*n,   //flags before compaction
           fs = f/s;  //flags after compaction        

// =========================================

 input  [k-v-i-1:0] rKey;
 input  [32-1:0] IV;
 input  set;
 input  [w*p-1:0] plainMem;

 output [fs-1:0] fd;
 output [v-1:0] vd;
 output done;

 output em;
 input  read;

 input  clk,pause,aclr;

 wire  [v-1:0] vKey,varKey;
 wire  rekey;
 wire  [w*n-1:0] keystream;
 wire  [n-1:0]  valid;
 wire  sync;
 reg   vsync,wsync;
 
 wire  [f-1:0] cmp;
 wire  [fs-1:0] flag;
 wire  read,wr;

// =========================================

  Del_vKeyGen    #(.v(v),.tap(28)) VGEN (set,{~set}, vKey,rekey,, clk,pause,aclr);
 `DEL_STREAM_CIPHER  #(.w(w))  SC0 ({ {(i){1'b0}},vKey,rKey },IV,rekey,pause, keystream,valid, clk,aclr);

 `DEL_CMP  #(.w(w),.p(p)) CMP0 (plainMem,keystream[0*w+w-1:0*w],valid[0], cmp[0*p+p-1:0*p],sync, clk,pause,aclr);

 `DEL_WR_VKEY_DELAY
  Del_vKeyGen    #(.v(v),.tap(28)) DGEN (set,vsync, varKey,,done, clk,pause,aclr);
 `DEL_WRCTRL  #(.p(p),.n(n),.s(s)) WRCTRL (cmp, flag,wr, clk,pause,aclr);

  FIFO_core32  BUFv (aclr,clk, varKey, read,wr,   ,,vd);

 genvar x;
 generate for (x=1; x<(fs/32); x=x+1) begin: BUF
  FIFO_core32  f  (aclr,clk, flag[(32*x)+32-1:(32*x)],  read,wr,   ,,fd[(32*x)+32-1:(32*x)]);
  end endgenerate
  FIFO_core32  BUFf0 (aclr,clk, flag[31:0],     read,wr, em,,fd[31:0]   );


// =========================================

endmodule





// ------------------------------------------------------
// ------------------------------------------------------






// *****************************************
//   Deluge - vKey Generator and Control
//   (for 0 < v =< 32)
//
// *****************************************

module Del_vKeyGen (set,start, vKey,rekey,done, clk,pause,aclr);
parameter v = 31 , tap = 28;

// =========================================

 input  set,start;

 output [v-1:0] vKey;
 output reg rekey;
 output reg done;

 input clk,pause,aclr;

 wire [v-1:0]  q;
 wire [1:0] check;
 reg  [2:0] state;
 reg  setx,startd,op;
 wire startx;

// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)     setx <= 'b0;
   else if(set) setx <= ~op;
   else         setx <= 'b0;
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)       startd <= 1'b0;
   else if(setx)  startd <= 1'b0;
   else if(start) startd <= 1'b1;
  end
 assign startx = start | startd;

 // - state machine for outputs rekey & done

 always @(posedge clk, posedge aclr) begin
   if(aclr) begin
       state <= 3'b000;
       rekey <= 1'b1;
       done  <= 1'b0;
       op <= 1'b1; 
      end
   else if(setx) begin
       state <= 3'b010;
       rekey <= 1'b0;
       done  <= 1'b0;
       op <= 1'b1;
      end
   else if(!pause) 
     casex({state,startx,{&check}})
       5'b000xx :  begin state <= 3'b001; rekey <= 1'b0; done <= 1'b0; op <= 1'b1; end

       5'b0010x :  begin state <= 3'b001; rekey <= 1'b0; done <= 1'b0; op <= 1'b1; end
       5'b0011x :  begin state <= 3'b011; rekey <= 1'b0; done <= 1'b1; op <= 1'b1; end

       5'b011xx :  begin state <= 3'b011; rekey <= 1'b0; done <= 1'b1; op <= 1'b0; end

       5'b0100x :  begin state <= 3'b010; rekey <= 1'b0; done <= 1'b0; op <= 1'b1; end
       5'b0101x :  begin state <= 3'b110; rekey <= 1'b1; done <= 1'b0; op <= 1'b1; end

       5'b110xx :  begin state <= 3'b111; rekey <= 1'b1; done <= 1'b0; op <= 1'b1; end

       5'b111x0 :  begin state <= 3'b111; rekey <= 1'b1; done <= 1'b0; op <= 1'b1; end
       5'b111x1 :  begin state <= 3'b101; rekey <= 1'b0; done <= 1'b1; op <= 1'b1; end

       5'b101xx :  begin state <= 3'b101; rekey <= 1'b0; done <= 1'b1; op <= 1'b0; end

     default begin state <= 3'b000; rekey <= 1'b1; done <= 1'b0; op <= 1'b1; end endcase
  end
  
  
 // - check vKey generation nearly finished

 generate if(v<16)
   Del_vKeyCheck1 #(v) C0 (setx,q, check, clk,pause,aclr);   //when v<16
 else
   Del_vKeyCheck2 #(v) CC0 (setx,q, check, clk,pause,aclr);  //when v>16 (max 32)
 endgenerate


 // - generate every possible vKey

 Del_LFSR #(v,tap) L0 (setx,startx, q, clk,pause,aclr);

 assign vKey = q;


// =========================================

endmodule



// *****************************************
//   vKey Finish Check (for v =< 16)
//
// *****************************************

module  Del_vKeyCheck1 (set,q, check, clk,pause,aclr);
parameter v = 9;

// =========================================

input set;
input [v-1:0] q;
output reg [1:0] check;

input clk,pause,aclr;

// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)        check <= 'b0;
   else if(set)    check <= 'b0;
   else if(!pause) check <= {&q[v-2:0],&q[v-2:0]};
  end 

// =========================================

endmodule



// *****************************************
//   vKey Finish Check (for 16 < v =< 32)
//
// *****************************************

module  Del_vKeyCheck2 (set,q, check, clk,pause,aclr);
parameter v = 31;

// =========================================

input set;
input [v-1:0] q;
output reg [1:0] check;

input clk,pause,aclr;

// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)        check <= 'b0;
   else if(set)    check <= 'b0;
   else if(!pause) check <= {&q[v-2:16],&q[15:0]};
  end 

// =========================================

endmodule



// *****************************************
//   LFSR   l bit
//
// *****************************************

module  Del_LFSR (set,start, q, clk,pause,aclr);
parameter l = 3, tap = 2;

// =========================================

input set,start;

output reg [l-1:0] q;

input clk,pause,aclr;

// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)        q <= {(l){1'b0}};
   else if(set)    q <= {(l){1'b1}};
   else if(start && !pause)  q <= { q[l-2:0], q[l-1] ^ q[tap-1] };
  end

// =========================================

endmodule




// ------------------------------------------------------






// *****************************************
//   Deluge - Px Partial Compare
//
// *****************************************

module Del_CMP_p (plain,keystream,valid, cmp,sync, clk,pause,aclr);
parameter w = 16, p = 64;

// =========================================

 input  [w*p-1:0] plain;
 input  [w-1:0]   keystream;
 input  valid;
 output [p-1:0] cmp;
 output sync;

 input clk,pause,aclr;

// =========================================

 genvar i;
 generate for (i=1; i<p; i=i+1) begin: C
  CMP_s2  #(.w(w)) p  (keystream,plain[w*i +w-1:w*i ],valid, cmp[i], , clk,pause,aclr);
  end endgenerate
  CMP_s2  #(.w(w)) C00 (keystream,plain[w*0 +w-1:w*0 ],valid, cmp[0],sync, clk,pause,aclr);


// =========================================

endmodule





// *****************************************
//   Compare - 2 stage
//
// *****************************************

module CMP_s2 (keystream,plainstream,k_valid, match,m_valid, clk,pause,aclr);
parameter w = 16,  //data width in compares
          c = 2,   //num of front cmp stages
          x = w/c; //front stage width

// =========================================

 input  [w-1:0]  keystream,plainstream;
 input  k_valid;
 output reg match;
 output reg m_valid;

 input clk,pause,aclr;

 reg   [c-1:0] cmp;
 reg   c_valid;

// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)        cmp <= 'b0;
   else if(!pause)
     if(!k_valid)  cmp <= 'b0;
     else begin
       cmp[1] <= &( keystream[1*x+x-1:1*x] ~^ plainstream[1*x+x-1:1*x] );
       cmp[0] <= &( keystream[0*x+x-1:0*x] ~^ plainstream[0*x+x-1:0*x] );
     end
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        c_valid <= 'b0;
   else if(!pause) c_valid <= k_valid;
  end


 always @(posedge clk, posedge aclr) begin
   if(aclr)         match <= 'b0;
   else if(!pause)   
     if(!c_valid)   match <= 'b0;
     else           match <= &cmp;  
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        m_valid <= 'b0;
   else if(!pause) m_valid <= c_valid;
  end


// =========================================

endmodule





// ------------------------------------------------------






// *****************************************
//   Deluge - Write Control Unit (3 stage)
//   
//   cmp max width on EP2CXX : 4096
//   includes compaction
//
// *****************************************

module Del_wrCTRL_s3c (cmp, flag,wr, clk,pause,aclr);
parameter p = 64, n = 2,
          s = 4,     //compaction factor
          f = p*n,   //flags before compaction
          fs = f/s;  //flags after compaction 

// =========================================

 input  [f-1:0] cmp;

 output reg [fs-1:0] flag;
 output reg wr;

 input clk,pause,aclr;

 wire [fs-1:0] signal;
 wire [256-1:0] sn_wr;

 reg  [fs-1:0] sign;
 wire [16-1:0] s_wr;

 genvar i;

// =========================================

 generate for (i=0; i<fs; i=i+1) begin: pack
   Del_OR #(.w(s)) wr (cmp[(i*s)+s-1:i*s], signal[i], clk,pause,aclr);
  end endgenerate
   
 generate for (i=0; i<(f/16); i=i+1) begin: wsn
   Del_OR #(.w(16)) wr (cmp[(i*16)+16-1:i*16], sn_wr[i], clk,pause,aclr);
  end endgenerate


 always @(posedge clk, posedge aclr) begin
   if(aclr)        sign <= 'b0;
   else if(!pause) sign <= signal;
  end
 
 generate for (i=0; i<16; i=i+1) begin: ws
   Del_OR #(.w(16)) wr (sn_wr[(i*16)+16-1:i*16], s_wr[i], clk,pause,aclr);
  end endgenerate


 always @(posedge clk, posedge aclr) begin
   if(aclr)        flag <= 'b0;
   else if(!pause) flag <= sign;
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        wr <= 'b0;
   else if(!pause) wr <= |s_wr;
  end


// =========================================

endmodule



// *****************************************
//   Deluge - Write Control Unit (2 stage)
//   
//   cmp max width on EP2CXX : 256
//   includes compaction
//
// *****************************************

module Del_wrCTRL_s2c (cmp, flag,wr, clk,pause,aclr);
parameter p = 64, n = 2,
          s = 2,     //compaction factor
          f = p*n,   //flags before compaction
          fs = f/s;  //flags after compaction 

// =========================================

 input  [f-1:0] cmp;

 output reg [fs-1:0] flag;
 output reg wr;

 input clk,pause,aclr;

 wire [fs-1:0] signal;
 wire [16-1:0] sn_wr;

 genvar i;

// =========================================

 generate for (i=0; i<fs; i=i+1) begin: pack
   Del_OR #(.w(s)) wr (cmp[(i*s)+s-1:i*s], signal[i], clk,pause,aclr);
  end endgenerate
   
 generate for (i=0; i<(f/16); i=i+1) begin: wsn
   Del_OR #(.w(16)) wr (cmp[(i*16)+16-1:i*16], sn_wr[i], clk,pause,aclr);
  end endgenerate


 always @(posedge clk, posedge aclr) begin
   if(aclr)        flag <= 'b0;
   else if(!pause) flag <= signal;
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        wr <= 'b0;
   else if(!pause) wr <= |sn_wr;
  end


// =========================================

endmodule



// *****************************************
//   Deluge - Write Control Unit (1 stage)
//   
//   cmp max width on EP2CXX : 16
//
// *****************************************

module Del_wrCTRL_s1 (cmp, flag,wr, clk,pause,aclr);
parameter p = 1, n = 8,
          s = 1,     //compaction factor
          f = p*n,   //flags before compaction
          fs = f/s;  //flags after compaction 

// =========================================

 input  [f-1:0] cmp;

 output reg [fs-1:0] flag;
 output reg wr;

 input clk,pause,aclr;

// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)        flag <= 'b0;
   else if(!pause) flag <= cmp;
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        wr <= 'b0;
   else if(!pause) wr <= |cmp;
  end


// =========================================

endmodule



// *****************************************
//   Deluge - OR function for wrCTRL
//
// *****************************************

module Del_OR (in, out, clk,pause,aclr);
parameter w = 16;

// =========================================

 input [w-1:0] in;
 output reg out;
 input clk,pause,aclr;

// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)        out <= 'b0;
   else if(!pause) out <= |in;
  end

// =========================================

endmodule





// ------------------------------------------------------
// ------------------------------------------------------








// *****************************************
//   Front-End Search Module    for n=2
//
// *****************************************

module Del_FrontEndSearch2 (rKey,IV,set,plainMem, fd,vd,done,em, read, clk,pause,aclr);
parameter  i = 1, j = 4, v = 31,
           h = 2, w = 16, k = 80,
           n = 2**i, p = 2**j, s = 2**h,
           f = p*n, fs = f/s;

// =========================================

 input  [k-v-i-1:0] rKey;
 input  [32-1:0] IV;
 input  set;
 input  [w*p-1:0] plainMem;

 output [fs-1:0] fd;
 output [v-1:0] vd;
 output done;

 output em;
 input  read;

 input  clk,pause,aclr;

 wire  [v-1:0] vKey,varKey;
 wire  rekey;
 wire  [w*n-1:0] keystream;
 wire  [n-1:0]  valid;
 wire  [f-1:0] cmp;
 wire  sync;
 reg   vsync,wsync;
 wire  read,wr;
 wire  [fs-1:0] flag;

// =========================================

  Del_vKeyGen    #(.v(v),.tap(28)) VGEN (set,{~set}, vKey,rekey,, clk,pause,aclr);
 `DEL_STREAM_CIPHER  #(.w(w))     SC0 ({ {(i){1'b0}},vKey,rKey },IV,rekey,pause, keystream,valid, clk,aclr);

 `DEL_CMP  #(.w(w),.p(p)) CMP1 (plainMem,keystream[1*w+w-1:1*w],valid[1], cmp[1*p+p-1:1*p], ,  clk,pause,aclr);
 `DEL_CMP  #(.w(w),.p(p)) CMP0 (plainMem,keystream[0*w+w-1:0*w],valid[0], cmp[0*p+p-1:0*p],sync, clk,pause,aclr);

 `DEL_WR_VKEY_DELAY
  Del_vKeyGen    #(.v(v),.tap(28)) DGEN (set,vsync, varKey,,done, clk,pause,aclr);
 `DEL_WRCTRL  #(.p(p),.n(n),.s(s)) WRCTRL (cmp, flag,wr, clk,pause,aclr);

  FIFO_core32  BUFv (aclr,clk, varKey, read,wr,   ,,vd);

 genvar x;
 generate for (x=1; x<(fs/32); x=x+1) begin: BUF
  FIFO_core32  f  (aclr,clk, flag[(32*x)+32-1:(32*x)],  read,wr,   ,,fd[(32*x)+32-1:(32*x)]);
  end endgenerate
  FIFO_core32  BUFf0 (aclr,clk, flag[31:0],     read,wr, em,,fd[31:0]   );


// =========================================

endmodule





// *****************************************
//   Front-End Search Module    for n=4
//
// *****************************************

module Del_FrontEndSearch4 (rKey,IV,set,plainMem, fd,vd,done,em, read, clk,pause,aclr);
parameter  i = 2, j = 4, v = 31,
           h = 2, w = 16, k = 80,
           n = 2**i, p = 2**j, s = 2**h,
           f = p*n, fs = f/s;

// =========================================

 input  [k-v-i-1:0] rKey;
 input  [32-1:0] IV;
 input  set;
 input  [w*p-1:0] plainMem;

 output [fs-1:0] fd;
 output [v-1:0] vd;
 output done;

 output em;
 input  read;

 input  clk,pause,aclr;

 wire  [v-1:0] vKey,varKey;
 wire  rekey;
 wire  [w*n-1:0] keystream;
 wire  [n-1:0]  valid;
 wire  [f-1:0] cmp;
 wire  sync;
 reg   vsync,wsync;
 wire  read,wr;
 wire  [fs-1:0] flag;

// =========================================

  Del_vKeyGen    #(.v(v),.tap(28)) VGEN (set,{~set}, vKey,rekey,, clk,pause,aclr);
 `DEL_STREAM_CIPHER  #(.w(w))    SC0 ({ {(i){1'b0}},vKey,rKey },IV,rekey,pause, keystream,valid, clk,aclr);

 `DEL_CMP  #(.w(w),.p(p)) CMP3 (plainMem,keystream[3*w+w-1:3*w],valid[3], cmp[3*p+p-1:3*p], ,  clk,pause,aclr);
 `DEL_CMP  #(.w(w),.p(p)) CMP2 (plainMem,keystream[2*w+w-1:2*w],valid[2], cmp[2*p+p-1:2*p], ,  clk,pause,aclr);
 `DEL_CMP  #(.w(w),.p(p)) CMP1 (plainMem,keystream[1*w+w-1:1*w],valid[1], cmp[1*p+p-1:1*p], ,  clk,pause,aclr);
 `DEL_CMP  #(.w(w),.p(p)) CMP0 (plainMem,keystream[0*w+w-1:0*w],valid[0], cmp[0*p+p-1:0*p],sync, clk,pause,aclr);

 `DEL_WR_VKEY_DELAY
  Del_vKeyGen    #(.v(v),.tap(28)) DGEN (set,vsync, varKey,,done, clk,pause,aclr);
 `DEL_WRCTRL  #(.p(p),.n(n),.s(s)) WRCTRL (cmp, flag,wr, clk,pause,aclr);

  FIFO_core32  BUFv (aclr,clk, varKey, read,wr,   ,,vd);

 genvar x;
 generate for (x=1; x<(fs/32); x=x+1) begin: BUF
  FIFO_core32  f  (aclr,clk, flag[(32*x)+32-1:(32*x)],  read,wr,   ,,fd[(32*x)+32-1:(32*x)]);
  end endgenerate
  FIFO_core32  BUFf0 (aclr,clk, flag[31:0],     read,wr, em,,fd[31:0]   );


// =========================================

endmodule





// *****************************************
//   Front-End Search Module    for 2x n=4
//   (max f=32)
//
// *****************************************

module Del_FrontEndSearch4x2 (rKey,IV,set,plainMem, fd,vd,done,em, read, clk,pause,aclr);
parameter  i = 3, j = 0, v = 31,
           h = 0, w = 16, k = 80,
           n = 2**i, p = 2**j, s = 2**h,
           f = p*n, fs = f/s;

// =========================================

 input  [k-v-i-1:0] rKey;
 input  [32-1:0] IV;
 input  set;
 input  [w*p-1:0] plainMem;

 output [fs-1:0] fd;
 output [v-1:0] vd;
 output done;

 output em;
 input  read;

 input  clk,pause,aclr;

 wire  [v-1:0] vKey,varKey;
 wire  rekey;
 wire  [w*n-1:0] keystream;
 wire  [n-1:0]  valid;
 wire  [f-1:0] cmp;
 wire  sync;
 reg   vsync,wsync;
 wire  read,wr;
 wire  [fs-1:0] flag;

// =========================================

  Del_vKeyGen    #(.v(v),.tap(28)) VGEN (set,{~set}, vKey,rekey,, clk,pause,aclr);

 `DEL_STREAM_CIPHER  #(.w(w))    SC1 ({ {(i-1){1'b0}},1'b1,vKey,rKey },IV,rekey,pause, keystream[w*n-1:w*n/2],valid[7:4], clk,aclr);

 `DEL_CMP  #(.w(w),.p(p)) CMP7 (plainMem,keystream[7*w+w-1:7*w],valid[7], cmp[7*p+p-1:7*p], ,  clk,pause,aclr);
 `DEL_CMP  #(.w(w),.p(p)) CMP6 (plainMem,keystream[6*w+w-1:6*w],valid[6], cmp[6*p+p-1:6*p], ,  clk,pause,aclr);
 `DEL_CMP  #(.w(w),.p(p)) CMP5 (plainMem,keystream[5*w+w-1:5*w],valid[5], cmp[5*p+p-1:5*p], ,  clk,pause,aclr);
 `DEL_CMP  #(.w(w),.p(p)) CMP4 (plainMem,keystream[4*w+w-1:4*w],valid[4], cmp[4*p+p-1:4*p], ,  clk,pause,aclr);

 `DEL_STREAM_CIPHER  #(.w(w))    SC0 ({ {(i-1){1'b0}},1'b0,vKey,rKey },IV,rekey,pause, keystream[(w*n/2)-1:0],valid[3:0], clk,aclr);

 `DEL_CMP  #(.w(w),.p(p)) CMP3 (plainMem,keystream[3*w+w-1:3*w],valid[3], cmp[3*p+p-1:3*p], ,  clk,pause,aclr);
 `DEL_CMP  #(.w(w),.p(p)) CMP2 (plainMem,keystream[2*w+w-1:2*w],valid[2], cmp[2*p+p-1:2*p], ,  clk,pause,aclr);
 `DEL_CMP  #(.w(w),.p(p)) CMP1 (plainMem,keystream[1*w+w-1:1*w],valid[1], cmp[1*p+p-1:1*p], ,  clk,pause,aclr);
 `DEL_CMP  #(.w(w),.p(p)) CMP0 (plainMem,keystream[0*w+w-1:0*w],valid[0], cmp[0*p+p-1:0*p],sync, clk,pause,aclr);

 `DEL_WR_VKEY_DELAY
  Del_vKeyGen    #(.v(v),.tap(28)) DGEN (set,vsync, varKey,,done, clk,pause,aclr);
 `DEL_WRCTRL  #(.p(p),.n(n),.s(s)) WRCTRL (cmp, flag,wr, clk,pause,aclr);

  FIFO_core32  BUFv (aclr,clk, varKey, read,wr,   ,,vd);

  FIFO_core32  BUFf0 (aclr,clk, flag[fs-1:0],     read,wr, em,,fd[fs-1:0]   );


// =========================================

endmodule





// *****************************************
//   Front-End Search Test Module
//   (n=2, short vKey cycle)
//   (Fully Tested)
// *****************************************

module Del_FrontEndSearchD9 (rKey,IV,set,plainMem, fd,vd,done,em, read, clk,pause,aclr);
parameter  i = 1, j = 4, v = 31,
           h = 2, w = 16, k = 80,
           n = 2**i, p = 2**j, s = 2**h,
           f = p*n, fs = f/s;

// =========================================

 input  [k-v-i-1:0] rKey;
 input  [32-1:0] IV;
 input  set;
 input  [w*p-1:0] plainMem;

 output [fs-1:0] fd;
 output [v-1:0] vd;
 output done;

 output em;
 input  read;

 input  clk,pause,aclr;

 wire  [9-1:0] vKey,varKey;
 wire  rekey;
 wire  [w*n-1:0] keystream;
 wire  [n-1:0]  valid;
 wire  [f-1:0] cmp;
 wire  sync;
 reg   vsync,wsync;
 wire  read,wr;
 wire  [fs-1:0] flag;

// =========================================

  Del_vKeyGen    #(.v(9),.tap(5)) VGEN (set,{~set}, vKey,rekey,, clk,pause,aclr);
 `DEL_STREAM_CIPHER  #(.w(w))   SC0 ({ {(i){1'b0}},{22'b0,vKey},rKey },IV,rekey,pause, keystream,valid, clk,aclr);

 `DEL_CMP  #(.w(w),.p(p)) CMP1 (plainMem,keystream[1*w+w-1:1*w],valid[1], cmp[1*p+p-1:1*p], ,  clk,pause,aclr);
 `DEL_CMP  #(.w(w),.p(p)) CMP0 (plainMem,keystream[0*w+w-1:0*w],valid[0], cmp[0*p+p-1:0*p],sync, clk,pause,aclr);

 `DEL_WR_VKEY_DELAY
  Del_vKeyGen    #(.v(9),.tap(5)) DGEN (set,vsync, varKey,,done, clk,pause,aclr);
 `DEL_WRCTRL  #(.p(p),.n(n),.s(s)) WRCTRL (cmp, flag,wr, clk,pause,aclr);

  FIFO_core32  BUFv (aclr,clk, {22'b0,varKey}, read,wr,   ,,vd);

 genvar x;
 generate for (x=1; x<(fs/32); x=x+1) begin: BUF
  FIFO_core32  f  (aclr,clk, flag[(32*x)+32-1:(32*x)],  read,wr,   ,,fd[(32*x)+32-1:(32*x)]);
  end endgenerate
  FIFO_core32  BUFf0 (aclr,clk, flag[31:0],     read,wr, em,,fd[31:0]   );


// =========================================

endmodule





// *****************************************
//   Front-End Search Test Module
//   (n=2, short vKey cycle)
//
// *****************************************

module Del_FrontEndSearchD28 (rKey,IV,set,plainMem, fd,vd,done,em, read, clk,pause,aclr);
parameter  i = 1, j = 4, v = 31,
           h = 2, w = 16, k = 80,
           n = 2**i, p = 2**j, s = 2**h,
           f = p*n, fs = f/s;

// =========================================

 input  [k-v-i-1:0] rKey;
 input  [32-1:0] IV;
 input  set;
 input  [w*p-1:0] plainMem;

 output [fs-1:0] fd;
 output [v-1:0] vd;
 output done;

 output em;
 input  read;

 input  clk,pause,aclr;

 wire  [28-1:0] vKey,varKey;
 wire  rekey;
 wire  [w*n-1:0] keystream;
 wire  [n-1:0]  valid;
 wire  [f-1:0] cmp;
 wire  sync;
 reg   vsync,wsync;
 wire  read,wr;
 wire  [fs-1:0] flag;

// =========================================

  Del_vKeyGen    #(.v(28),.tap(25)) VGEN (set,{~set}, vKey,rekey,, clk,pause,aclr);
 `DEL_STREAM_CIPHER  #(.w(w))   SC0 ({ {(i){1'b0}},{3'b0,vKey},rKey },IV,rekey,pause, keystream,valid, clk,aclr);

 `DEL_CMP  #(.w(w),.p(p)) CMP1 (plainMem,keystream[1*w+w-1:1*w],valid[1], cmp[1*p+p-1:1*p], ,  clk,pause,aclr);
 `DEL_CMP  #(.w(w),.p(p)) CMP0 (plainMem,keystream[0*w+w-1:0*w],valid[0], cmp[0*p+p-1:0*p],sync, clk,pause,aclr);

 `DEL_WR_VKEY_DELAY
  Del_vKeyGen    #(.v(28),.tap(25)) DGEN (set,vsync, varKey,,done, clk,pause,aclr);
 `DEL_WRCTRL  #(.p(p),.n(n),.s(s)) WRCTRL (cmp, flag,wr, clk,pause,aclr);

  FIFO_core32  BUFv (aclr,clk, {3'b0,varKey}, read,wr,   ,,vd);

 genvar x;
 generate for (x=1; x<(fs/32); x=x+1) begin: BUF
  FIFO_core32  f  (aclr,clk, flag[(32*x)+32-1:(32*x)],  read,wr,   ,,fd[(32*x)+32-1:(32*x)]);
  end endgenerate
  FIFO_core32  BUFf0 (aclr,clk, flag[31:0],     read,wr, em,,fd[31:0]   );


// =========================================

endmodule





// *****************************************
//   Front-End Search Test Module
//   (n=4, short vKey cycle)
//
// *****************************************

module Del_FrontEndSearchT9 (rKey,IV,set,plainMem, fd,vd,done,em, read, clk,pause,aclr);
parameter  i = 2, j = 4, v = 31,
           h = 2, w = 16, k = 80,
           n = 2**i, p = 2**j, s = 2**h,
           f = p*n, fs = f/s;

// =========================================

 input  [k-v-i-1:0] rKey;
 input  [32-1:0] IV;
 input  set;
 input  [w*p-1:0] plainMem;

 output [fs-1:0] fd;
 output [v-1:0] vd;
 output done;

 output em;
 input  read;

 input  clk,pause,aclr;

 wire  [9-1:0] vKey,varKey;
 wire  rekey;
 wire  [w*n-1:0] keystream;
 wire  [n-1:0]  valid;
 wire  [f-1:0] cmp;
 wire  sync;
 reg   vsync,wsync;
 wire  read,wr;
 wire  [fs-1:0] flag;

// =========================================

  Del_vKeyGen  #(.v(9),.tap(5)) VGEN (set,{~set}, vKey,rekey,, clk,pause,aclr);
 `DEL_STREAM_CIPHER  #(.w(w))   SC0 ({ {(i){1'b0}},{22'b0,vKey},rKey },IV,rekey,pause, keystream,valid, clk,aclr);

 `DEL_CMP  #(.w(w),.p(p)) CMP3 (plainMem,keystream[3*w+w-1:3*w],valid[3], cmp[3*p+p-1:3*p], ,  clk,pause,aclr);
 `DEL_CMP  #(.w(w),.p(p)) CMP2 (plainMem,keystream[2*w+w-1:2*w],valid[2], cmp[2*p+p-1:2*p], ,  clk,pause,aclr);
 `DEL_CMP  #(.w(w),.p(p)) CMP1 (plainMem,keystream[1*w+w-1:1*w],valid[1], cmp[1*p+p-1:1*p], ,  clk,pause,aclr);
 `DEL_CMP  #(.w(w),.p(p)) CMP0 (plainMem,keystream[0*w+w-1:0*w],valid[0], cmp[0*p+p-1:0*p],sync, clk,pause,aclr);

 `DEL_WR_VKEY_DELAY
  Del_vKeyGen  #(.v(9),.tap(5)) DGEN (set,vsync, varKey,,done, clk,pause,aclr);
 `DEL_WRCTRL  #(.p(p),.n(n),.s(s)) WRCTRL (cmp, flag,wr, clk,pause,aclr);

  FIFO_core32  BUFv (aclr,clk, {22'b0,varKey}, read,wr,   ,,vd);

 genvar x;
 generate for (x=1; x<(fs/32); x=x+1) begin: BUF
  FIFO_core32  f  (aclr,clk, flag[(32*x)+32-1:(32*x)],  read,wr,   ,,fd[(32*x)+32-1:(32*x)]);
  end endgenerate
  FIFO_core32  BUFf0 (aclr,clk, flag[31:0],     read,wr, em,,fd[31:0]   );


// =========================================

endmodule