
// *****************************************
//   DELUGE - Translation Unit
//
//   ver:1.2
//   status: under test   Mar/07
//   notes: (allows case where i=0 or j=0)
//
// *****************************************
//   Verilog Code (30 Mar 07)
//   Copyright Iain Devlin 2006-07
//
// *****************************************




// *****************************************
//   Deluge - Translation Unit
//
// *****************************************

module Del_TranslationUnit (df,dv,empty, pid,id,vKey,rd,em, read, clk,pause,aclr);
parameter i = 2, j = 4, v = 31,
          f = 64;
`ifdef DEL_NO_ID parameter nid = 1; `else parameter nid = 0; `endif
`ifdef DEL_NO_PID parameter npid = 1; `else parameter npid = 0; `endif

// =========================================

 input  [f-1:0] df;
 input  [v-1:0] dv;
 input  empty;
 
 output [j+npid-1:0] pid;
 output [i+nid-1:0] id;
 output [v-1:0] vKey;
 output rd;
 output em;

 input  read;
 input  clk,pause,aclr;

 wire  [j+i-1:0] jd,qj;
 wire  wr;

// =========================================

 Del_FlagEncoder  #(.f(f),.j(j+i)) FE  (empty,df, rd,jd,wr, clk,pause,aclr);

 FIFO_core32  BUFj (aclr,clk, jd, read,wr, ,,qj );
 FIFO_core32  BUFv (aclr,clk, dv, read,wr, em,,vKey);

 assign id = qj[j+i-1:j-nid];
 assign pid = qj[j+npid-1:0];


// =========================================

endmodule





// *****************************************
//   Flag Encoder
//
// *****************************************

module Del_FlagEncoder (empty,flag, rd,jd,wr, clk,pause,aclr);
parameter f=1024, j=10;
//parameter f=256, j=8;
//parameter f=64, j=6;

// =========================================

 input empty;
 input [f-1:0] flag;

 output reg rd;
 output wire [j-1:0] jd;
 output wire wr;
 
 input clk,pause,aclr;

 reg  setx,set;
 wire done;

// =========================================

 `DEL_FLAG_ENCODE  FE0 (set,set,flag[f-1:0], jd[j-1:0],wr,done, clk,pause,aclr);
 //Del_FlagEncode_1024T10 FE0 (set,set,flag[f-1:0], jd[j-1:0],wr,done, clk,pause,aclr);
 //Del_FlagEncode_256T8 FE0 (set,set,flag[f-1:0], jd[j-1:0],wr,done, clk,pause,aclr);
 //Del_FlagEncode_64T6 FE0 (set,set,flag[f-1:0], jd[j-1:0],wr,done, clk,pause,aclr);

 always @(posedge clk, posedge aclr) begin
   if(aclr)        rd <= 'b0;
   else if(!pause) rd <= done & ~|{rd,set,empty};
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        setx <= 'b0;
   else if(!pause) setx <= rd;
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        set <= 'b0;
   else if(!pause) set <= rd | setx;
  end


// =========================================

endmodule





// *****************************************
//   Flag 1024 to 10 encoder
//
// *****************************************

module Del_FlagEncode_1024T10(setup,iset,in, out,valid,done, clk,pause,aclr);
parameter f=1024, j=10;

// =========================================

 input setup,iset;
 input [f-1:0] in;

 output reg [j-1:0] out;
 output reg valid,done;

 input clk,pause,aclr;

 reg  [4-1:0] set;
 wire [4-1:0] val,dn;
 wire [4*(j-2)-1:0] c;

// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)       set <= 'b0;
   else if(setup) set <= 4'b1111;
   else if(!pause) begin
     set[3] <= iset;
     set[2] <= |{iset,~dn[3]};
     set[1] <= |{iset,~dn[3],~dn[2]};
     set[0] <= |{iset,~dn[3],~dn[2],~dn[1]};
    end
  end

 Del_FlagEncode_256T8 FE3 (setup,set[3],in[1023:768], c[31:24],val[3],dn[3], clk,pause,aclr );
 Del_FlagEncode_256T8 FE2 (setup,set[2],in[767:512],  c[23:16],val[2],dn[2], clk,pause,aclr );
 Del_FlagEncode_256T8 FE1 (setup,set[1],in[511:256],  c[15:8], val[1],dn[1], clk,pause,aclr );
 Del_FlagEncode_256T8 FE0 (setup,set[0],in[255:0],    c[7:0],  val[0],dn[0], clk,pause,aclr );

 always @(posedge clk, posedge aclr) begin
   if(aclr)      out <= 'b0;
   else if(!pause) begin
     casex(dn)
       4'b0xxx : out <= {2'b11,c[31:24]};
       4'b10xx : out <= {2'b10,c[23:16]};
       4'b110x : out <= {2'b01,c[15:8]};
       4'b1110 : out <= {2'b00,c[7:0]};
       4'b1111 : out <= 'b0;
       default   out <= 'b0;
     endcase
    end
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        valid <= 'b0;
   else if(setup)  valid <= 'b0;
   else if(!pause) valid <= |val;
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        done <= 1'b1;
   else if(setup)  done <= 1'b0;
   else if(!pause) done <= &dn;
  end
 

// =========================================

endmodule





// *****************************************
//   Flag 256 to 8 encoder
//
// *****************************************

module Del_FlagEncode_256T8(setup,iset,in, out,valid,done, clk,pause,aclr);
parameter f=256, j=8;

// =========================================

 input setup,iset;
 input [f-1:0] in;

 output reg [j-1:0] out;
 output reg valid,done;

 input clk,pause,aclr;

 reg  [4-1:0] set;
 wire [4-1:0] val,dn;
 wire [4*(j-2)-1:0] c;

// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)       set <= 'b0;
   else if(setup) set <= 4'b1111;
   else if(!pause) begin
     set[3] <= iset;
     set[2] <= |{iset,~dn[3]};
     set[1] <= |{iset,~dn[3],~dn[2]};
     set[0] <= |{iset,~dn[3],~dn[2],~dn[1]};
    end
  end

 Del_FlagEncode_64T6 FE3 (setup,set[3],in[255:192], c[23:18],val[3],dn[3], clk,pause,aclr );
 Del_FlagEncode_64T6 FE2 (setup,set[2],in[191:128], c[17:12],val[2],dn[2], clk,pause,aclr );
 Del_FlagEncode_64T6 FE1 (setup,set[1],in[127:64],  c[11:6], val[1],dn[1], clk,pause,aclr );
 Del_FlagEncode_64T6 FE0 (setup,set[0],in[63:0],    c[5:0],  val[0],dn[0], clk,pause,aclr );

 always @(posedge clk, posedge aclr) begin
   if(aclr)      out <= 'b0;
   else if(!pause) begin
     casex(dn)
       4'b0xxx : out <= {2'b11,c[23:18]};
       4'b10xx : out <= {2'b10,c[17:12]};
       4'b110x : out <= {2'b01,c[11:6]};
       4'b1110 : out <= {2'b00,c[5:0]};
       4'b1111 : out <= 'b0;
       default   out <= 'b0;
     endcase
    end
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        valid <= 'b0;
   else if(setup)  valid <= 'b0;
   else if(!pause) valid <= |val;
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        done <= 1'b1;
   else if(setup)  done <= 1'b0;
   else if(!pause) done <= &dn;
  end
 

// =========================================

endmodule




// *****************************************
//   Flag 64 to 6 encoder
//
// *****************************************

module Del_FlagEncode_64T6(setup,iset,in, out,valid,done, clk,pause,aclr);
parameter f=64, j=6;

// =========================================

 input setup,iset;
 input [f-1:0] in;

 output reg [j-1:0] out;
 output reg valid,done;

 input clk,pause,aclr;

 reg  [4-1:0] set;
 wire [4-1:0] val,dn;
 wire [4*(j-2)-1:0] c;

// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)       set <= 'b0;
   else if(setup) set <= 4'b1111;
   else if(!pause) begin
     set[3] <= iset;
     set[2] <= |{iset,~dn[3]};
     set[1] <= |{iset,~dn[3],~dn[2]};
     set[0] <= |{iset,~dn[3],~dn[2],~dn[1]};
    end
  end

 Del_FlagEncode_16T4 FE3 (setup,set[3],in[63:48], c[15:12],val[3],dn[3], clk,pause,aclr );
 Del_FlagEncode_16T4 FE2 (setup,set[2],in[47:32], c[11:8], val[2],dn[2], clk,pause,aclr );
 Del_FlagEncode_16T4 FE1 (setup,set[1],in[31:16], c[7:4],  val[1],dn[1], clk,pause,aclr );
 Del_FlagEncode_16T4 FE0 (setup,set[0],in[15:0],  c[3:0],  val[0],dn[0], clk,pause,aclr );

 always @(posedge clk, posedge aclr) begin
   if(aclr)      out <= 'b0;
   else if(!pause) begin
     casex(dn)
       4'b0xxx : out <= {2'b11,c[15:12]};
       4'b10xx : out <= {2'b10,c[11:8]};
       4'b110x : out <= {2'b01,c[7:4]};
       4'b1110 : out <= {2'b00,c[3:0]};
       4'b1111 : out <= 'b0;
       default   out <= 'b0;
     endcase
    end
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        valid <= 'b0;
   else if(setup)  valid <= 'b0;
   else if(!pause) valid <= |val;
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        done <= 1'b1;
   else if(setup)  done <= 1'b0;
   else if(!pause) done <= &dn;
  end
 

// =========================================

endmodule





// *****************************************
//   Flag 16 to 4 encoder
//
// *****************************************

module Del_FlagEncode_16T4(setup,iset,in, out,valid,done, clk,pause,aclr);
parameter f=16, j=4;

// =========================================

 input setup,iset;
 input [f-1:0] in;

 output reg [j-1:0] out;
 output reg valid,done;

 input clk,pause,aclr;

 reg  [4-1:0] set;
 wire [4-1:0] val,dn;
 wire [4*(j-2)-1:0] c;



// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)       set <= 'b0;
   else if(setup) set <= 4'b1111;
   else if(!pause) begin
     set[3] <= iset;
     set[2] <= |{iset,~dn[3]};
     set[1] <= |{iset,~dn[3],~dn[2]};
     set[0] <= |{iset,~dn[3],~dn[2],~dn[1]};
    end
  end

 Del_FlagEncode_4T2 FE3 (setup,set[3],in[15:12], c[7:6],val[3],dn[3], clk,pause,aclr );
 Del_FlagEncode_4T2 FE2 (setup,set[2],in[11:8],  c[5:4],val[2],dn[2], clk,pause,aclr );
 Del_FlagEncode_4T2 FE1 (setup,set[1],in[7:4],   c[3:2],val[1],dn[1], clk,pause,aclr );
 Del_FlagEncode_4T2 FE0 (setup,set[0],in[3:0],   c[1:0],val[0],dn[0], clk,pause,aclr );

 always @(posedge clk, posedge aclr) begin
   if(aclr)      out <= 'b0;
   else if(!pause) begin
     casex(dn)
       4'b0xxx : out <= {2'b11,c[7:6]};
       4'b10xx : out <= {2'b10,c[5:4]};
       4'b110x : out <= {2'b01,c[3:2]};
       4'b1110 : out <= {2'b00,c[1:0]};
       4'b1111 : out <= 'b0;
       default   out <= 'b0;
     endcase
    end
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        valid <= 'b0;
   else if(setup)  valid <= 'b0;
   else if(!pause) valid <= |val;
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        done <= 1'b1;
   else if(setup)  done <= 1'b0;
   else if(!pause) done <= &dn;
  end
 

// =========================================

endmodule





// *****************************************
//   Flag 4 to 2 encoder
//
// *****************************************

module Del_FlagEncode_4T2(setup,set,in, out,wr,done, clk,pause,aclr);
parameter f=4,j=2;

// =========================================

 input setup,set;
 input [f-1:0] in;

 output reg [j-1:0] out;
 output reg wr, done;

 input clk,pause,aclr;

 reg  [f-1:0] mask;

// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)     mask <= 'b0;
   else if(set) mask <= {(f){1'b1}};
   else if(!pause) begin
     case({ (in[3]&mask[3])|(in[2]&mask[2]) , (in[3]&mask[3])|((~(in[2]&mask[2]))&(in[1]&mask[1])) })
       2'b11 :  mask <= 4'b0111;
       2'b10 :  mask <= 4'b0011;
       2'b01 :  mask <= 4'b0001;
       2'b00 :  mask <= 4'b0000;
     endcase
    end
  end

  always @(posedge clk, posedge aclr) begin
   if(aclr)        wr <= 'b0;
   else if(set)    wr <= 'b0;
   else if(!pause) wr <= |(in & mask);
  end

  always @(posedge clk, posedge aclr) begin
   if(aclr)        done <= 1'b1;
   else if(setup)  done <= 1'b0;
   else if(!pause) done <= ~|(in & mask);
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        out <= 'b0;
   else if(set)    out <= 'b0;
   else if(!pause) out <= {(in[3] & mask[3])|(in[2]&mask[2]) , (in[3]&mask[3])|((~(in[2]&mask[2]))&(in[1]&mask[1])) };
  end


// =========================================

endmodule






// ------------------------------------------------------
// ------------------------------------------------------







// *****************************************
//   Deluge - Translation Unit for i=0
//
// *****************************************

module Del_TranslationUnit0 (df,dv,empty, pid,id,vKey,rd,em, read, clk,pause,aclr);
parameter i = 0, j = 4, v = 31,
          f = 64;

// =========================================

 input  [f-1:0] df;
 input  [v-1:0] dv;
 input  empty;
 
 output [j-1:0] pid;
 output [i-1:0] id;
 output [v-1:0] vKey;
 output rd;
 output em;

 input  read;
 input  clk,pause,aclr;
 
 wire  [j-1:0] jd,qj;
 wire  wr;

// =========================================

 Del_FlagEncoder  #(.f(f),.j(j+i)) FE  (empty,df, rd,jd,wr, clk,pause,aclr);

 FIFO_core32  BUFj (aclr,clk, jd, read,wr, ,,qj );
 FIFO_core32  BUFv (aclr,clk, dv, read,wr, em,,vKey);

 assign pid = qj[j-1:0];


// =========================================

endmodule