
// *******************************************
//   Deluge
//   A Key Search System for Stream Ciphers
//
//   ver:1.11
//   status: under test   Mar/07
//   notes: (new dual quad mode, improved performance vs ver.0)
//
// *******************************************
//   Verilog Code (30 Mar 07)
//   Copyright Iain Devlin 2006-07
//
// *******************************************



`define PLAIN_ADDR_WIDTH 9  //addr space relating to plainstream numbers
//max   9 = 512 flags per cipher
//      8 = 256
//      7 = 128
//min   6 = 64  
//brute 0 = 1

//`define DEL_NO_PID  true  //for brute force without data 

//---------------------------------

`define DLG_N2  true   //choose front end setup
`define DLG_T256 true  //choose translation setup
`define DLG_S2  true   //choose back end setup

//---------------------------------

`ifdef DLG_N4x2
 `define DEL_FRONTEND_SEARCH Del_FrontEndSearch4x2
 `define SC_ADDR_WIDTH  3
 `define DEL_STREAM_CIPHER Grain1X_n4
`endif

`ifdef DLG_N4
 `define DEL_FRONTEND_SEARCH Del_FrontEndSearch4
 `define SC_ADDR_WIDTH  2
 `define DEL_STREAM_CIPHER Grain1X_n4
`endif

`ifdef DLG_N2
 `define DEL_FRONTEND_SEARCH Del_FrontEndSearch2
 `define SC_ADDR_WIDTH  1
 `define DEL_STREAM_CIPHER Grain1X_n2
`endif

`ifdef DLG_N
 `define DEL_FRONTEND_SEARCH Del_FrontEndSearch
 `define SC_ADDR_WIDTH  0
 `define DEL_NO_ID  true
 `define DEL_STREAM_CIPHER Grain1p_16
`endif

`ifdef DLG_N2s
 `define DEL_FRONTEND_SEARCH Del_FrontEndSearchD28
 `define SC_ADDR_WIDTH  1
 `define DEL_STREAM_CIPHER Grain1X_n2
`endif

`ifdef DLG_N4TEST
 `define DEL_FRONTEND_SEARCH Del_FrontEndSearchT9
 `define SC_ADDR_WIDTH  2
 `define DEL_STREAM_CIPHER Grain1X_n4
`endif

`ifdef DLG_N2TEST
 `define DEL_FRONTEND_SEARCH Del_FrontEndSearchD9
 `define SC_ADDR_WIDTH  1
 `define DEL_STREAM_CIPHER Grain1X_n2
`endif

//---------------------------------

`ifdef DLG_T1024
 `define DEL_FLAG_ENCODE Del_FlagEncode_1024T10
`endif
`ifdef DLG_T256
 `define DEL_FLAG_ENCODE Del_FlagEncode_256T8
`endif
`ifdef DLG_T64
 `define DEL_FLAG_ENCODE Del_FlagEncode_64T6
`endif
`ifdef DLG_T16
 `define DEL_FLAG_ENCODE Del_FlagEncode_16T4
`endif

//---------------------------------

`ifdef DLG_S8
 `define COMPACTION_LEVEL  3
 `define DEL_BACKEND_SEARCH Del_BackEndFullSearch8
`endif

`ifdef DLG_S4
 `define COMPACTION_LEVEL  2
 `define DEL_BACKEND_SEARCH Del_BackEndFullSearch4
`endif

`ifdef DLG_S2
 `define COMPACTION_LEVEL  1
 `define DEL_BACKEND_SEARCH Del_BackEndFullSearch2
`endif

`ifdef DLG_S
 `define COMPACTION_LEVEL  0
 `define DEL_BACKEND_SEARCH Del_BackEndFullSearch
`endif

//---------------------------------

`define DLG_vKEY_SIZE 31   //size of variable key space tested by chip before done
`define DLG_DATA_WIDTH 16  //width of data bus, max 31

//---------------------------------

`define SC_KEY_SIZE  80    //key size of target cipher
`define SC_WIDTH  16       //width of cipher keystream output

// Grain1X_n4    //4 Grain v1 pipelines merged, 13.6kLE
// Grain1X_n2    //2 Grain v1 pipelines merged, 7.6kLE
// Grain1p_16    //Grain v1 pipelined, keystream width 16, 4.7kLE
// Grain128X_n2r //2 Grain-128 pipelines merged, output width 16, 7.5kLE
// Grain128p_32r //Grain-128 pipelined, keystream 32, output width 16
// TriviumP64    //Trivium pipelined, keystream width 64, 8.5kLE    

//Del_FrontEndSearch4   //full vKey cycle for n=4
//Del_FrontEndSearch2   //full vKey cycle for n=2
//Del_FrontEndSearch    //full (2^V)-1 vKey cycle
//Del_FrontEndSearchT9  //test setup for debug, (2^9)-1 vKey cycle, n=4
//Del_FrontEndSearchD9  //test setup for debug, (2^9)-1 vKey cycle, n=2
//Del_FrontEndSearchD28 //test setup for debug, (2^28)-1 vKey cycle, n=2

//---------------------------------

//Del_FlagEncode_1024T10   //max 1024 compacted flags,long delay, large
//Del_FlagEncode_256T8     //max 256 compacted flags, standard, 1.5kLE + 0.3kLE for FIFOs
//Del_FlagEncode_64T6      //max 64 compacted flags, short delay, 0.4kLE + 0.1kLE for FIFOs
//Del_FlagEncode_16T4      //max 16 compacted flags, min delay, small

//---------------------------------

//`define COMPACTION_LEVEL 2   //h = 0,1,2,3

//Del_BackEndFullSearch8  //compaction level 3, 8 memory search
//Del_BackEndFullSearch4  //compaction level 2, 4 memory search
//Del_BackEndFullSearch2  //compaction level 1, 2 memory search
//Del_BackEndFullSearch   //for use when no compaction

//choose cipher to match frontend
// Grain1_16
// Grain128_16
// Trivium_64   

//---------------------------------







// *******************************************
//   Deluge - Chip Level Module
//
// *******************************************

module Deluge (_data,_addr,_cs,_wIr,_gpause, clk,_aclr, xclk,_xclr);
parameter d = `DLG_DATA_WIDTH,
          i = `SC_ADDR_WIDTH, j = `PLAIN_ADDR_WIDTH, v = `DLG_vKEY_SIZE,
          h = `COMPACTION_LEVEL,
          w = `SC_WIDTH,
          k = `SC_KEY_SIZE,
          r = k-v-i,    //size of register key part
          p = 2**j,     //number of plainstreams
          a = j+3,      //plainstream addr space, assumes each split in 5
          wp = w*p;

`ifdef DEL_NO_PID  parameter npid = 1; `else parameter npid = 0; `endif
 
// ===========================================

 inout  [d-1:0] _data;
 input  [a+npid+2-1:0] _addr;
 input  _cs,_wIr;
 input  _gpause;
 
 input  clk,xclk;
 input  _aclr,_xclr;

 wire  gpause;
 wire  aclr,xaclr,xclr;

 wire  [r-1:0]  rKey;
 wire  [32-1:0] IV;
 wire  [a+npid-1:0]  xadr;
 wire  [4-1:0]  xwr;
 wire  [0  :0]  wr;
 wire  [d-1:0]  xdataout,xdatain,datain;

 wire  [wp-1:0] plainreg;
 wire  [w-1:0] mdatain;
 wire  [a-1:0] madr;
 wire  mwr;

 wire  pause;
 wire  rekey;
 wire  keyflag,xkeyflag;
 wire  done,xdone;

 wire  [d-1:0] xdataY,xdataK;
 wire  [80-1:0] oKey;

 // - - - - - - - - - - - - - - - - - - - - - -
 // for system debug add
 // - - - - - - - - - - - - - - - - - - - - - -
 // * module head: , plainstream0,plainstream1,pid,fem,bem);
 // - - - - - - - - - - - - - - - - - - - - - -
 // * output fem,bem;
 // * output [15:0] plainstream0,plainstream1;
 // * output [j-h-1:0] pid;
 // - - - - - - - - - - - - - - - - - - - - - -

 // - - - - - - - - - - - - - - - - - - - - - -
 // for hardware test add
 // - - - - - - - - - - - - - - - - - - - - - -
 // * module head: , _d88);
 // - - - - - - - - - - - - - - - - - - - - - -
 // wire [j-h-1:0] pid;
 // reg  [7:0] dpid;
 // wire [d-1:0] d88;
 // output [d-1:0] _d88;
 // - - - - - - - - - - - - - - - - - - - - - -
 // ** 7 SEGMENT DISPLAY **
 // always @(posedge clk) begin
 //   if(aclr)      dpid <= 'b0;
 //   else dpid[j-h-1:0] <= pid;
 //  end
 // CTRL_BinTo7Seg D80 (1'b0, dpid[7:4], d88[15:8], clk,aclr);
 // CTRL_BinTo7Seg D08 (1'b0, dpid[3:0], d88[7:0] , clk,aclr);
 // assign _d88 = ~d88;
 // - - - - - - - - - - - - - - - - - - - - - -
 // ** LEDs **
 // * module head: D0,D1,D2);
 // * output reg D0,D1,D2;
 // always @* begin
 //   D0 <= aclr;
 //   D1 <= xaclr;
 //   D2 <= xclr;
 //  end
 // - - - - - - - - - - - - - - - - - - - - - -


// ===========================================


// *** DOMAIN TRANSLATION EXTERNAL ***

 // * i/o head: input  _clk,xclk; wire clk;
 // * PLL clkdiv (_clk,clk);
 // * assign clk = xclk;

 // - - - - - - - - - - - - - - - - - - - - -
 // - synchronise inputs (avoid metastability)
 // - - - - - - - - - - - - - - - - - - - - -
  DLG_sync_unit  Dclr  (_aclr,   aclr,   clk, 1'b0);
  DLG_sync_unit  Dclrx (_aclr,   xaclr, xclk, 1'b0);
  DLG_sync_unit  Dxclr (_xclr,   xclr,  xclk, 1'b0);
  DLG_sync_unit  Dglbp (_gpause, gpause, clk, aclr);
 // - - - - - - - - - - - - - - - - - - - - -


// *** EXTERNAL CLOCK DOMAIN ***

 DLG_Interface #(.d(d),.a(a+npid)) DLI (_data,_addr,_cs,_wIr,xdataout, xdatain,xadr,xwr, xclk,xaclr);

 DLG_iReg  #(.w(r),.d(d))  DLkey (xdatain,xadr[1:0],xwr[1], rKey, xclk,xclr);
 DLG_iReg  #(.w(32),.d(d)) DLiv  (xdatain,xadr[1:0],xwr[2], IV,   xclk,xclr);

 DLG_ipReg  #(.wp(wp),.d(d),.j(j+npid))  DLplain  (xdatain,xadr[a+npid-1:3],{xwr[3]&(~|xadr[2:0])}, plainreg, xclk,xclr);
 DLG_ipMem  #(.d(d),.a(a))  DLpMem  (xdatain,xadr[a-1:0],xwr[3],{2'b0}, mdatain,madr,mwr, xclk,xaclr);

 DLG_sMonitor #(.d(d))       DLsMON (xkeyflag,xdone, xdataY, xclk,xaclr);
 DLG_oReg     #(.w(k),.d(d)) DLoKey (oKey,xadr[3:1],xkeyflag, xdataK, xclk,xaclr);

 DLG_OutMUX  #(.d(d))  DLoMUX (xdataY,xdataK,xadr[0], xdataout);



// *** DOMAIN TRANSLATION INTERNAL ***

 // - - - - - - - - - - - - - - - - - - - - -
 // synchronise between clk domains
 // - - - - - - - - - - - - - - - - - - - - -
  DLG_sync_unit #(d) Ddin (xdatain, datain, clk,aclr);
  DLG_sync_unit      Dwr  (xwr[0],  wr[0],  clk,aclr);
  DLG_sync_unit  Dflg  (keyflag, xkeyflag, xclk,xaclr);
  DLG_sync_unit  Ddn   (done,    xdone,    xclk,xaclr);
 // - - - - - - - - - - - - - - - - - - - - -
 // NOTE  plainreg,rKey,IV  have no syncing
 //   and so require time for stablisation
 //   on completetion of data loading.
 //   (no oKey syncing due to keyflag sync)
 // - - - - - - - - - - - - - - - - - - - - -
 
 

// ***INTERNAL CLOCK DOMAIN *** 

 Del_Core  #(.i(i),.j(j),.v(v),.h(h),.w(w),.k(k)) 
   DLCORE (rKey,IV,rekey,plainreg,mdatain,madr,mwr, oKey,keyflag,done, clk,pause,aclr, xclk);
 // - - - - - - - - - - - - - - - - - - - - - -
 // for hardware test add: ,pid);
 // for system debug add: ,fem,bem,plainstream0,plainstream1,pid);
 // - - - - - - - - - - - - - - - - - - - - - -

 DLG_sControl  #(.d(d)) DLsCTRL (keyflag,done,datain,wr[0], pause,rekey, clk,gpause,aclr);



// ===========================================

endmodule





// ------------------------------------------------------
// ------------------------------------------------------







// *******************************************
//   2reg Synchronise Unit
//   (avoids metastability at interfaces)
//
// *******************************************

 module DLG_sync_unit (in, out, clk,aclr);
 parameter width = 1;

// ===========================================

 input [width-1:0] in;
 output reg [width-1:0] out;
 input clk,aclr;

 reg [width-1:0] sync;

// ===========================================

 always@(posedge clk, posedge aclr) begin
   if(aclr) sync <= 'b0;
   else     sync <= in;
  end

 always@(posedge clk, posedge aclr) begin
   if(aclr) out <= 'b0;
   else     out <= sync;
  end

// ===========================================

endmodule





// ------------------------------------------------------
// ------------------------------------------------------






// *******************************************
//   Deluge Core Interface Unit
//
// *******************************************

module  DLG_Interface (_bus,_addr,_cs,_wIr, dout,din,adr,wr, clk,aclr);
parameter d = 16, a = 7;

// ===========================================

 inout  [d-1:0] _bus;

 input  [a+2-1:0] _addr;
 input  _cs,_wIr;

 input  [d-1:0] dout; 
 output reg [d-1:0] din;
 
 output reg [a-1:0] adr;
 output reg [4-1:0] wr;

 input  clk,aclr;

 reg  oe;

 wire  [d-1:0] ibus;
 wire  [a+2-1:0] addr;
 wire  cs,wIr;

// ===========================================


// ** SYNCHRONISE INPUTS **
// - avoid metastability

 DLG_sync_unit         DIcs   (_cs,   cs,   clk,aclr);
 DLG_sync_unit         DIwr   (_wIr,  wIr,  clk,aclr);
 DLG_sync_unit  #(d)   DIbus  (_bus,  ibus, clk,aclr);
 DLG_sync_unit  #(a+2) DIaddr (_addr, addr, clk,aclr);


// ** TRISTATED DATA BUS ** 
 
 always@* begin
   if(aclr)       oe <= 'b0;
   else if(!_cs)  oe <= 'b0;  // non sync'd to avoid bus conflict
   else if(!_wIr) oe <= 1'b1; // output dout
   else           oe <= 'b0;  // output Z
  end   
 assign _bus = (oe ? dout : {d{1'bz}});	

 always@(posedge clk, posedge aclr) begin
   if(aclr)      din <= 'b0;
   else if(!cs)  din <= 'b0;
   else if(!wIr) din <= 'b0;
   else          din <= ibus;
 end


// ** ADDRESS DECODER **
// - top 2 bits give wr
// - remaining give adr

 always@(posedge clk, posedge aclr) begin
   if(aclr)      wr <= 'b0;
   else if(!cs)  wr <= 'b0;
   else if(!wIr) wr <= 'b0;
   else
     case(addr[a+2-1:a])
      2'h0 : wr <= 4'b0001;  // write ack
      2'h1 : wr <= 4'b0010;  // write rKey
      2'h2 : wr <= 4'b0100;  // write IV
      2'h3 : wr <= 4'b1000;  // write pMem
     default  wr <= 'b0;
     endcase
 end
 
 always@(posedge clk, posedge aclr) begin
   if(aclr)     adr <= 'b0;
   else if(!cs) adr <= 'b0;  
   else         adr <= addr[a-1:0]; 
  end


// ===========================================

endmodule





// *******************************************
//   Deluge Input Register  (2bit addr max)
//
// *******************************************

module  DLG_iReg (data,addr,wr, out, clk,aclr);
parameter w = 32, d = 16;

// ===========================================

 input [d-1:0] data;
 input [2-1:0] addr;
 input  wr;
 
 output [w-1:0] out;

 input  clk,aclr;

 reg  [4*d-1:0] datareg;

// ===========================================


// - load reg in parts

 always@(posedge clk, posedge aclr) begin
   if(aclr)     datareg <= 'b0;
   else if(wr) begin
     case(addr)
      2'h0 : datareg[0*d+d-1:0*d] <= data;
      2'h1 : datareg[1*d+d-1:1*d] <= data;
      2'h2 : datareg[2*d+d-1:2*d] <= data;
      2'h3 : datareg[3*d+d-1:3*d] <= data;
     default  datareg <= 'b0;
     endcase
   end
 end
 
 
// - constantly output full reg contents

 assign out = datareg[w-1:0];


// ===========================================

endmodule





// *******************************************
//   Deluge PlainMem Register  (max d=4w)
//
// *******************************************

module  DLG_ipMem  (data,addr,wr,sel, mdata,madr,mwr, clk,aclr);
parameter d = 64, w = 16, a = 6+3;

// ===========================================

 input [d-1:0] data;
 input [a-1:0] addr;
 input  wr;
 input [1:0] sel;

 output [w-1:0] mdata;
 output reg [a-1:0] madr;
 output reg  mwr;
 
 input  clk,aclr;

 reg [4*d-1:0] mreg;

// ===========================================


// ** TRANSFER MEMORY COMMAND SIGNALS **
// - load data in parts if (d < w)

 always@(posedge clk, posedge aclr) begin
   if(aclr)  mreg <= 'b0;
   else case(sel)
      2'h0 : mreg[0*d+d-1:0*d] <= data;
      2'h1 : mreg[1*d+d-1:1*d] <= data;
      2'h2 : mreg[2*d+d-1:2*d] <= data;
      2'h3 : mreg[3*d+d-1:3*d] <= data;
     default  mreg <= 'b0;
     endcase
  end
 assign mdata = mreg[w-1:0];

 always@(posedge clk, posedge aclr) begin
   if(aclr)  madr <= 'b0;
   else      madr <= addr;
  end
 
 always@(posedge clk, posedge aclr) begin
   if(aclr)  mwr <= 'b0;
   else      mwr <= wr;
  end


// ===========================================

endmodule





// *******************************************
//   Deluge Input Plain Reg   (512 enrty max)
//
//   must be a better way of coding this!
//   or maybe there isn't ...
// 
// *******************************************
//
module  DLG_ipReg (data,addr,wr, out, clk,aclr);
parameter wp = 64, d = 16, j = 4,
          x = 9, max = 2**x;

// ===========================================

 input [d-1:0] data;
 input [j-1:0] addr;
 input  wr;
 
 output [wp-1:0] out;

 input  clk,aclr;

 wire [x-1:0] address;
 reg  [max*d-1:0] datareg;
 
// ===========================================

 assign address = { {(x-j){1'b0}},addr };


// - load reg in parts

 always@(posedge clk, posedge aclr) begin
   if(aclr)     datareg <= 'b0;
   else if(wr) begin
     case(address)
     
      9'd0 : datareg[0*d+d-1:0*d] <= data;
      9'd1 : datareg[1*d+d-1:1*d] <= data;
      9'd2 : datareg[2*d+d-1:2*d] <= data;
      9'd3 : datareg[3*d+d-1:3*d] <= data;

      9'd4 : datareg[4*d+d-1:4*d] <= data;
      9'd5 : datareg[5*d+d-1:5*d] <= data;
      9'd6 : datareg[6*d+d-1:6*d] <= data;
      9'd7 : datareg[7*d+d-1:7*d] <= data;

      9'd8  : datareg[ 8*d+d-1: 8*d] <= data;
      9'd9  : datareg[ 9*d+d-1: 9*d] <= data;
      9'd10 : datareg[10*d+d-1:10*d] <= data;
      9'd11 : datareg[11*d+d-1:11*d] <= data;
      9'd12 : datareg[12*d+d-1:12*d] <= data;
      9'd13 : datareg[13*d+d-1:13*d] <= data;
      9'd14 : datareg[14*d+d-1:14*d] <= data;
      9'd15 : datareg[15*d+d-1:15*d] <= data;

      9'd16 : datareg[16*d+d-1:16*d] <= data;
      9'd17 : datareg[17*d+d-1:17*d] <= data;
      9'd18 : datareg[18*d+d-1:18*d] <= data;
      9'd19 : datareg[19*d+d-1:19*d] <= data;
      9'd20 : datareg[20*d+d-1:20*d] <= data;
      9'd21 : datareg[21*d+d-1:21*d] <= data;
      9'd22 : datareg[22*d+d-1:22*d] <= data;
      9'd23 : datareg[23*d+d-1:23*d] <= data;

      9'd24 : datareg[24*d+d-1:24*d] <= data;
      9'd25 : datareg[25*d+d-1:25*d] <= data;
      9'd26 : datareg[26*d+d-1:26*d] <= data;
      9'd27 : datareg[27*d+d-1:27*d] <= data;
      9'd28 : datareg[28*d+d-1:28*d] <= data;
      9'd29 : datareg[29*d+d-1:29*d] <= data;
      9'd30 : datareg[30*d+d-1:30*d] <= data;
      9'd31 : datareg[31*d+d-1:31*d] <= data;


      9'd32 : datareg[32*d+d-1:32*d] <= data;
      9'd33 : datareg[33*d+d-1:33*d] <= data;
      9'd34 : datareg[34*d+d-1:34*d] <= data;
      9'd35 : datareg[35*d+d-1:35*d] <= data;
      9'd36 : datareg[36*d+d-1:36*d] <= data;
      9'd37 : datareg[37*d+d-1:37*d] <= data;
      9'd38 : datareg[38*d+d-1:38*d] <= data;
      9'd39 : datareg[39*d+d-1:39*d] <= data;

      9'd40 : datareg[40*d+d-1:40*d] <= data;
      9'd41 : datareg[41*d+d-1:41*d] <= data;
      9'd42 : datareg[42*d+d-1:42*d] <= data;
      9'd43 : datareg[43*d+d-1:43*d] <= data;
      9'd44 : datareg[44*d+d-1:44*d] <= data;
      9'd45 : datareg[45*d+d-1:45*d] <= data;
      9'd46 : datareg[46*d+d-1:46*d] <= data;
      9'd47 : datareg[47*d+d-1:47*d] <= data;     

      9'd48 : datareg[48*d+d-1:48*d] <= data;
      9'd49 : datareg[49*d+d-1:49*d] <= data;
      9'd50 : datareg[50*d+d-1:50*d] <= data;
      9'd51 : datareg[51*d+d-1:51*d] <= data;
      9'd52 : datareg[52*d+d-1:52*d] <= data;
      9'd53 : datareg[53*d+d-1:53*d] <= data;
      9'd54 : datareg[54*d+d-1:54*d] <= data;
      9'd55 : datareg[55*d+d-1:55*d] <= data;

      9'd56 : datareg[56*d+d-1:56*d] <= data;
      9'd57 : datareg[57*d+d-1:57*d] <= data;
      9'd58 : datareg[58*d+d-1:58*d] <= data;
      9'd59 : datareg[59*d+d-1:59*d] <= data;
      9'd60 : datareg[60*d+d-1:60*d] <= data;
      9'd61 : datareg[61*d+d-1:61*d] <= data;
      9'd62 : datareg[62*d+d-1:62*d] <= data;
      9'd63 : datareg[63*d+d-1:63*d] <= data;

     
      9'd64 : datareg[ 64*d+d-1: 64*d] <= data;
      9'd65 : datareg[ 65*d+d-1: 65*d] <= data;
      9'd66 : datareg[ 66*d+d-1: 66*d] <= data;
      9'd67 : datareg[ 67*d+d-1: 67*d] <= data;
      9'd68 : datareg[ 68*d+d-1: 68*d] <= data;
      9'd69 : datareg[ 69*d+d-1: 69*d] <= data;
      9'd70 : datareg[ 70*d+d-1: 70*d] <= data;
      9'd71 : datareg[ 71*d+d-1: 71*d] <= data;

      9'd72 : datareg[ 72*d+d-1: 72*d] <= data;
      9'd73 : datareg[ 73*d+d-1: 73*d] <= data;
      9'd74 : datareg[ 74*d+d-1: 74*d] <= data;
      9'd75 : datareg[ 75*d+d-1: 75*d] <= data;
      9'd76 : datareg[ 76*d+d-1: 76*d] <= data;
      9'd77 : datareg[ 77*d+d-1: 77*d] <= data;
      9'd78 : datareg[ 78*d+d-1: 78*d] <= data;
      9'd79 : datareg[ 79*d+d-1: 79*d] <= data;

      9'd80 : datareg[ 80*d+d-1: 80*d] <= data;
      9'd81 : datareg[ 81*d+d-1: 81*d] <= data;
      9'd82 : datareg[ 82*d+d-1: 82*d] <= data;
      9'd83 : datareg[ 83*d+d-1: 83*d] <= data;
      9'd84 : datareg[ 84*d+d-1: 84*d] <= data;
      9'd85 : datareg[ 85*d+d-1: 85*d] <= data;
      9'd86 : datareg[ 86*d+d-1: 86*d] <= data;
      9'd87 : datareg[ 87*d+d-1: 87*d] <= data;

      9'd88 : datareg[ 88*d+d-1: 88*d] <= data;
      9'd89 : datareg[ 89*d+d-1: 89*d] <= data;
      9'd90 : datareg[ 90*d+d-1: 90*d] <= data;
      9'd91 : datareg[ 91*d+d-1: 91*d] <= data;
      9'd92 : datareg[ 92*d+d-1: 92*d] <= data;
      9'd93 : datareg[ 93*d+d-1: 93*d] <= data;
      9'd94 : datareg[ 94*d+d-1: 94*d] <= data;
      9'd95 : datareg[ 95*d+d-1: 95*d] <= data;


      9'd96  : datareg[ 96*d+d-1: 96*d] <= data;
      9'd97  : datareg[ 97*d+d-1: 97*d] <= data;
      9'd98  : datareg[ 98*d+d-1: 98*d] <= data;
      9'd99  : datareg[ 99*d+d-1: 99*d] <= data;
      9'd100 : datareg[100*d+d-1:100*d] <= data;
      9'd101 : datareg[101*d+d-1:101*d] <= data;
      9'd102 : datareg[102*d+d-1:102*d] <= data;
      9'd103 : datareg[103*d+d-1:103*d] <= data;

      9'd104 : datareg[104*d+d-1:104*d] <= data;
      9'd105 : datareg[105*d+d-1:105*d] <= data;
      9'd106 : datareg[106*d+d-1:106*d] <= data;
      9'd107 : datareg[107*d+d-1:107*d] <= data;
      9'd108 : datareg[108*d+d-1:108*d] <= data;
      9'd109 : datareg[109*d+d-1:109*d] <= data;
      9'd110 : datareg[110*d+d-1:110*d] <= data;
      9'd111 : datareg[111*d+d-1:111*d] <= data; 

      9'd112 : datareg[112*d+d-1:112*d] <= data;
      9'd113 : datareg[113*d+d-1:113*d] <= data;
      9'd114 : datareg[114*d+d-1:114*d] <= data;
      9'd115 : datareg[115*d+d-1:115*d] <= data;
      9'd116 : datareg[116*d+d-1:116*d] <= data;
      9'd117 : datareg[117*d+d-1:117*d] <= data;
      9'd118 : datareg[118*d+d-1:118*d] <= data;
      9'd119 : datareg[119*d+d-1:119*d] <= data;

      9'd120 : datareg[120*d+d-1:120*d] <= data;
      9'd121 : datareg[121*d+d-1:121*d] <= data;
      9'd122 : datareg[122*d+d-1:122*d] <= data;
      9'd123 : datareg[123*d+d-1:123*d] <= data;
      9'd124 : datareg[124*d+d-1:124*d] <= data;
      9'd125 : datareg[125*d+d-1:125*d] <= data;
      9'd126 : datareg[126*d+d-1:126*d] <= data;
      9'd127 : datareg[127*d+d-1:127*d] <= data;



      9'd128 : datareg[128*d+d-1:128*d] <= data;
      9'd129 : datareg[129*d+d-1:129*d] <= data;
      9'd130 : datareg[130*d+d-1:130*d] <= data;
      9'd131 : datareg[131*d+d-1:131*d] <= data;
      9'd132 : datareg[132*d+d-1:132*d] <= data;
      9'd133 : datareg[133*d+d-1:133*d] <= data;
      9'd134 : datareg[134*d+d-1:134*d] <= data;
      9'd135 : datareg[135*d+d-1:135*d] <= data;

      9'd136 : datareg[136*d+d-1:136*d] <= data;
      9'd137 : datareg[137*d+d-1:137*d] <= data;
      9'd138 : datareg[138*d+d-1:138*d] <= data;
      9'd139 : datareg[139*d+d-1:139*d] <= data;
      9'd140 : datareg[140*d+d-1:140*d] <= data;
      9'd141 : datareg[141*d+d-1:141*d] <= data;
      9'd142 : datareg[142*d+d-1:142*d] <= data;
      9'd143 : datareg[143*d+d-1:143*d] <= data;

      9'd144 : datareg[144*d+d-1:144*d] <= data;
      9'd145 : datareg[145*d+d-1:145*d] <= data;
      9'd146 : datareg[146*d+d-1:146*d] <= data;
      9'd147 : datareg[147*d+d-1:147*d] <= data;
      9'd148 : datareg[148*d+d-1:148*d] <= data;
      9'd149 : datareg[149*d+d-1:149*d] <= data;
      9'd150 : datareg[150*d+d-1:150*d] <= data;
      9'd151 : datareg[151*d+d-1:151*d] <= data;

      9'd152 : datareg[152*d+d-1:152*d] <= data;
      9'd153 : datareg[153*d+d-1:153*d] <= data;
      9'd154 : datareg[154*d+d-1:154*d] <= data;
      9'd155 : datareg[155*d+d-1:155*d] <= data;
      9'd156 : datareg[156*d+d-1:156*d] <= data;
      9'd157 : datareg[157*d+d-1:157*d] <= data;
      9'd158 : datareg[158*d+d-1:158*d] <= data;
      9'd159 : datareg[159*d+d-1:159*d] <= data;


      9'd160 : datareg[160*d+d-1:160*d] <= data;
      9'd161 : datareg[161*d+d-1:161*d] <= data;
      9'd162 : datareg[162*d+d-1:162*d] <= data;
      9'd163 : datareg[163*d+d-1:163*d] <= data;
      9'd164 : datareg[164*d+d-1:164*d] <= data;
      9'd165 : datareg[165*d+d-1:165*d] <= data;
      9'd166 : datareg[166*d+d-1:166*d] <= data;
      9'd167 : datareg[167*d+d-1:167*d] <= data;

      9'd168 : datareg[168*d+d-1:168*d] <= data;
      9'd169 : datareg[169*d+d-1:169*d] <= data;
      9'd170 : datareg[170*d+d-1:170*d] <= data;
      9'd171 : datareg[171*d+d-1:171*d] <= data;
      9'd172 : datareg[172*d+d-1:172*d] <= data;
      9'd173 : datareg[173*d+d-1:173*d] <= data;
      9'd174 : datareg[174*d+d-1:174*d] <= data;
      9'd175 : datareg[175*d+d-1:175*d] <= data;

      9'd176 : datareg[176*d+d-1:176*d] <= data;
      9'd177 : datareg[177*d+d-1:177*d] <= data;
      9'd178 : datareg[178*d+d-1:178*d] <= data;
      9'd179 : datareg[179*d+d-1:179*d] <= data;
      9'd180 : datareg[180*d+d-1:180*d] <= data;
      9'd181 : datareg[181*d+d-1:181*d] <= data;
      9'd182 : datareg[182*d+d-1:182*d] <= data;
      9'd183 : datareg[183*d+d-1:183*d] <= data;

      9'd184 : datareg[184*d+d-1:184*d] <= data;
      9'd185 : datareg[185*d+d-1:185*d] <= data;
      9'd186 : datareg[186*d+d-1:186*d] <= data;
      9'd187 : datareg[187*d+d-1:187*d] <= data;
      9'd188 : datareg[188*d+d-1:188*d] <= data;
      9'd189 : datareg[189*d+d-1:189*d] <= data;
      9'd190 : datareg[190*d+d-1:190*d] <= data;
      9'd191 : datareg[191*d+d-1:191*d] <= data;


      9'd192 : datareg[192*d+d-1:192*d] <= data;
      9'd193 : datareg[193*d+d-1:193*d] <= data;
      9'd194 : datareg[194*d+d-1:194*d] <= data;
      9'd195 : datareg[195*d+d-1:195*d] <= data;
      9'd196 : datareg[196*d+d-1:196*d] <= data;
      9'd197 : datareg[197*d+d-1:197*d] <= data;
      9'd198 : datareg[198*d+d-1:198*d] <= data;
      9'd199 : datareg[199*d+d-1:199*d] <= data;

      9'd200 : datareg[200*d+d-1:200*d] <= data;
      9'd201 : datareg[201*d+d-1:201*d] <= data;
      9'd202 : datareg[202*d+d-1:202*d] <= data;
      9'd203 : datareg[203*d+d-1:203*d] <= data;
      9'd204 : datareg[204*d+d-1:204*d] <= data;
      9'd205 : datareg[205*d+d-1:205*d] <= data;
      9'd206 : datareg[206*d+d-1:206*d] <= data;
      9'd207 : datareg[207*d+d-1:207*d] <= data;

      9'd208 : datareg[208*d+d-1:208*d] <= data;
      9'd209 : datareg[209*d+d-1:209*d] <= data;
      9'd210 : datareg[210*d+d-1:210*d] <= data;
      9'd211 : datareg[211*d+d-1:211*d] <= data;
      9'd212 : datareg[212*d+d-1:212*d] <= data;
      9'd213 : datareg[213*d+d-1:213*d] <= data;
      9'd214 : datareg[214*d+d-1:214*d] <= data;
      9'd215 : datareg[215*d+d-1:215*d] <= data;

      9'd216 : datareg[216*d+d-1:216*d] <= data;
      9'd217 : datareg[217*d+d-1:217*d] <= data;
      9'd218 : datareg[218*d+d-1:218*d] <= data;
      9'd219 : datareg[219*d+d-1:219*d] <= data;
      9'd220 : datareg[220*d+d-1:220*d] <= data;
      9'd221 : datareg[221*d+d-1:221*d] <= data;
      9'd222 : datareg[222*d+d-1:222*d] <= data;
      9'd223 : datareg[223*d+d-1:223*d] <= data;


      9'd224 : datareg[224*d+d-1:224*d] <= data;
      9'd225 : datareg[225*d+d-1:225*d] <= data;
      9'd226 : datareg[226*d+d-1:226*d] <= data;
      9'd227 : datareg[227*d+d-1:227*d] <= data;
      9'd228 : datareg[228*d+d-1:228*d] <= data;
      9'd229 : datareg[229*d+d-1:229*d] <= data;
      9'd230 : datareg[230*d+d-1:230*d] <= data;
      9'd231 : datareg[231*d+d-1:231*d] <= data;

      9'd232 : datareg[232*d+d-1:232*d] <= data;
      9'd233 : datareg[233*d+d-1:233*d] <= data;
      9'd234 : datareg[234*d+d-1:234*d] <= data;
      9'd235 : datareg[235*d+d-1:235*d] <= data;
      9'd236 : datareg[236*d+d-1:236*d] <= data;
      9'd237 : datareg[237*d+d-1:237*d] <= data;
      9'd238 : datareg[238*d+d-1:238*d] <= data;
      9'd239 : datareg[239*d+d-1:239*d] <= data;

      9'd240 : datareg[240*d+d-1:240*d] <= data;
      9'd241 : datareg[241*d+d-1:241*d] <= data;
      9'd242 : datareg[242*d+d-1:242*d] <= data;
      9'd243 : datareg[243*d+d-1:243*d] <= data;
      9'd244 : datareg[244*d+d-1:244*d] <= data;
      9'd245 : datareg[245*d+d-1:245*d] <= data;
      9'd246 : datareg[246*d+d-1:246*d] <= data;
      9'd247 : datareg[247*d+d-1:247*d] <= data;

      9'd248 : datareg[248*d+d-1:248*d] <= data;
      9'd249 : datareg[249*d+d-1:249*d] <= data;
      9'd250 : datareg[250*d+d-1:250*d] <= data;
      9'd251 : datareg[251*d+d-1:251*d] <= data;
      9'd252 : datareg[252*d+d-1:252*d] <= data;
      9'd253 : datareg[253*d+d-1:253*d] <= data;
      9'd254 : datareg[254*d+d-1:254*d] <= data;
      9'd255 : datareg[255*d+d-1:255*d] <= data;



      9'd256 : datareg[256*d+d-1:256*d] <= data;
      9'd257 : datareg[257*d+d-1:257*d] <= data;
      9'd258 : datareg[258*d+d-1:258*d] <= data;
      9'd259 : datareg[259*d+d-1:259*d] <= data;
      9'd260 : datareg[260*d+d-1:260*d] <= data;
      9'd261 : datareg[261*d+d-1:261*d] <= data;
      9'd262 : datareg[262*d+d-1:262*d] <= data;
      9'd263 : datareg[263*d+d-1:263*d] <= data;

      9'd264 : datareg[264*d+d-1:264*d] <= data;
      9'd265 : datareg[265*d+d-1:265*d] <= data;
      9'd266 : datareg[266*d+d-1:266*d] <= data;
      9'd267 : datareg[267*d+d-1:267*d] <= data;
      9'd268 : datareg[268*d+d-1:268*d] <= data;
      9'd269 : datareg[269*d+d-1:269*d] <= data;
      9'd270 : datareg[270*d+d-1:270*d] <= data;
      9'd271 : datareg[271*d+d-1:271*d] <= data;

      9'd272 : datareg[272*d+d-1:272*d] <= data;
      9'd273 : datareg[273*d+d-1:273*d] <= data;
      9'd274 : datareg[274*d+d-1:274*d] <= data;
      9'd275 : datareg[275*d+d-1:275*d] <= data;
      9'd276 : datareg[276*d+d-1:276*d] <= data;
      9'd277 : datareg[277*d+d-1:277*d] <= data;
      9'd278 : datareg[278*d+d-1:278*d] <= data;
      9'd279 : datareg[279*d+d-1:279*d] <= data;

      9'd280 : datareg[280*d+d-1:280*d] <= data;
      9'd281 : datareg[281*d+d-1:281*d] <= data;
      9'd282 : datareg[282*d+d-1:282*d] <= data;
      9'd283 : datareg[283*d+d-1:283*d] <= data;
      9'd284 : datareg[284*d+d-1:284*d] <= data;
      9'd285 : datareg[285*d+d-1:285*d] <= data;
      9'd286 : datareg[286*d+d-1:286*d] <= data;
      9'd287 : datareg[287*d+d-1:287*d] <= data;


      9'd288 : datareg[288*d+d-1:288*d] <= data;
      9'd289 : datareg[289*d+d-1:289*d] <= data;
      9'd290 : datareg[290*d+d-1:290*d] <= data;
      9'd291 : datareg[291*d+d-1:291*d] <= data;
      9'd292 : datareg[292*d+d-1:292*d] <= data;
      9'd293 : datareg[293*d+d-1:293*d] <= data;
      9'd294 : datareg[294*d+d-1:294*d] <= data;
      9'd295 : datareg[295*d+d-1:295*d] <= data;

      9'd296 : datareg[296*d+d-1:296*d] <= data;
      9'd297 : datareg[297*d+d-1:297*d] <= data;
      9'd298 : datareg[298*d+d-1:298*d] <= data;
      9'd299 : datareg[299*d+d-1:299*d] <= data;
      9'd300 : datareg[300*d+d-1:300*d] <= data;
      9'd301 : datareg[301*d+d-1:301*d] <= data;
      9'd302 : datareg[302*d+d-1:302*d] <= data;
      9'd303 : datareg[303*d+d-1:303*d] <= data;

      9'd304 : datareg[304*d+d-1:304*d] <= data;
      9'd305 : datareg[305*d+d-1:305*d] <= data;
      9'd306 : datareg[306*d+d-1:306*d] <= data;
      9'd307 : datareg[307*d+d-1:307*d] <= data;
      9'd308 : datareg[308*d+d-1:308*d] <= data;
      9'd309 : datareg[309*d+d-1:309*d] <= data;
      9'd310 : datareg[310*d+d-1:310*d] <= data;
      9'd311 : datareg[311*d+d-1:311*d] <= data;

      9'd312 : datareg[312*d+d-1:312*d] <= data;
      9'd313 : datareg[313*d+d-1:313*d] <= data;
      9'd314 : datareg[314*d+d-1:314*d] <= data;
      9'd315 : datareg[315*d+d-1:315*d] <= data;
      9'd316 : datareg[316*d+d-1:316*d] <= data;
      9'd317 : datareg[317*d+d-1:317*d] <= data;
      9'd318 : datareg[318*d+d-1:318*d] <= data;
      9'd319 : datareg[319*d+d-1:319*d] <= data;


      9'd320 : datareg[320*d+d-1:320*d] <= data;
      9'd321 : datareg[321*d+d-1:321*d] <= data;
      9'd322 : datareg[322*d+d-1:322*d] <= data;
      9'd323 : datareg[323*d+d-1:323*d] <= data;
      9'd324 : datareg[324*d+d-1:324*d] <= data;
      9'd325 : datareg[325*d+d-1:325*d] <= data;
      9'd326 : datareg[326*d+d-1:326*d] <= data;
      9'd327 : datareg[327*d+d-1:327*d] <= data;

      9'd328 : datareg[328*d+d-1:328*d] <= data;
      9'd329 : datareg[329*d+d-1:329*d] <= data;
      9'd330 : datareg[330*d+d-1:330*d] <= data;
      9'd331 : datareg[331*d+d-1:331*d] <= data;
      9'd332 : datareg[332*d+d-1:332*d] <= data;
      9'd333 : datareg[333*d+d-1:333*d] <= data;
      9'd334 : datareg[334*d+d-1:334*d] <= data;
      9'd335 : datareg[335*d+d-1:335*d] <= data;

      9'd336 : datareg[336*d+d-1:336*d] <= data;
      9'd337 : datareg[337*d+d-1:337*d] <= data;
      9'd338 : datareg[338*d+d-1:338*d] <= data;
      9'd339 : datareg[339*d+d-1:339*d] <= data;
      9'd340 : datareg[340*d+d-1:340*d] <= data;
      9'd341 : datareg[341*d+d-1:341*d] <= data;
      9'd342 : datareg[342*d+d-1:342*d] <= data;
      9'd343 : datareg[343*d+d-1:343*d] <= data;

      9'd344 : datareg[344*d+d-1:344*d] <= data;
      9'd345 : datareg[345*d+d-1:345*d] <= data;
      9'd346 : datareg[346*d+d-1:346*d] <= data;
      9'd347 : datareg[347*d+d-1:347*d] <= data;
      9'd348 : datareg[348*d+d-1:348*d] <= data;
      9'd349 : datareg[349*d+d-1:349*d] <= data;
      9'd350 : datareg[350*d+d-1:350*d] <= data;
      9'd351 : datareg[351*d+d-1:351*d] <= data;


      9'd352 : datareg[352*d+d-1:352*d] <= data;
      9'd353 : datareg[353*d+d-1:353*d] <= data;
      9'd354 : datareg[354*d+d-1:354*d] <= data;
      9'd355 : datareg[355*d+d-1:355*d] <= data;
      9'd356 : datareg[356*d+d-1:356*d] <= data;
      9'd357 : datareg[357*d+d-1:357*d] <= data;
      9'd358 : datareg[358*d+d-1:358*d] <= data;
      9'd359 : datareg[359*d+d-1:359*d] <= data;

      9'd360 : datareg[360*d+d-1:360*d] <= data;
      9'd361 : datareg[361*d+d-1:361*d] <= data;
      9'd362 : datareg[362*d+d-1:362*d] <= data;
      9'd363 : datareg[363*d+d-1:363*d] <= data;
      9'd364 : datareg[364*d+d-1:364*d] <= data;
      9'd365 : datareg[365*d+d-1:365*d] <= data;
      9'd366 : datareg[366*d+d-1:366*d] <= data;
      9'd367 : datareg[367*d+d-1:367*d] <= data;

      9'd368 : datareg[368*d+d-1:368*d] <= data;
      9'd369 : datareg[369*d+d-1:369*d] <= data;
      9'd370 : datareg[370*d+d-1:370*d] <= data;
      9'd371 : datareg[371*d+d-1:371*d] <= data;
      9'd372 : datareg[372*d+d-1:372*d] <= data;
      9'd373 : datareg[373*d+d-1:373*d] <= data;
      9'd374 : datareg[374*d+d-1:374*d] <= data;
      9'd375 : datareg[375*d+d-1:375*d] <= data;

      9'd376 : datareg[376*d+d-1:376*d] <= data;
      9'd377 : datareg[377*d+d-1:377*d] <= data;
      9'd378 : datareg[378*d+d-1:378*d] <= data;
      9'd379 : datareg[379*d+d-1:379*d] <= data;
      9'd380 : datareg[380*d+d-1:380*d] <= data;
      9'd381 : datareg[381*d+d-1:381*d] <= data;
      9'd382 : datareg[382*d+d-1:382*d] <= data;
      9'd383 : datareg[383*d+d-1:383*d] <= data;



      9'd384 : datareg[384*d+d-1:384*d] <= data;
      9'd385 : datareg[385*d+d-1:385*d] <= data;
      9'd386 : datareg[386*d+d-1:386*d] <= data;
      9'd387 : datareg[387*d+d-1:387*d] <= data;
      9'd388 : datareg[388*d+d-1:388*d] <= data;
      9'd389 : datareg[389*d+d-1:389*d] <= data;
      9'd390 : datareg[390*d+d-1:390*d] <= data;
      9'd391 : datareg[391*d+d-1:391*d] <= data;

      9'd392 : datareg[392*d+d-1:392*d] <= data;
      9'd393 : datareg[393*d+d-1:393*d] <= data;
      9'd394 : datareg[394*d+d-1:394*d] <= data;
      9'd395 : datareg[395*d+d-1:395*d] <= data;
      9'd396 : datareg[396*d+d-1:396*d] <= data;
      9'd397 : datareg[397*d+d-1:397*d] <= data;
      9'd398 : datareg[398*d+d-1:398*d] <= data;
      9'd399 : datareg[399*d+d-1:399*d] <= data;

      9'd400 : datareg[400*d+d-1:400*d] <= data;
      9'd401 : datareg[401*d+d-1:401*d] <= data;
      9'd402 : datareg[402*d+d-1:402*d] <= data;
      9'd403 : datareg[403*d+d-1:403*d] <= data;
      9'd404 : datareg[404*d+d-1:404*d] <= data;
      9'd405 : datareg[405*d+d-1:405*d] <= data;
      9'd406 : datareg[406*d+d-1:406*d] <= data;
      9'd407 : datareg[407*d+d-1:407*d] <= data;

      9'd408 : datareg[408*d+d-1:408*d] <= data;
      9'd409 : datareg[409*d+d-1:409*d] <= data;
      9'd410 : datareg[410*d+d-1:410*d] <= data;
      9'd411 : datareg[411*d+d-1:411*d] <= data;
      9'd412 : datareg[412*d+d-1:412*d] <= data;
      9'd413 : datareg[413*d+d-1:413*d] <= data;
      9'd414 : datareg[414*d+d-1:414*d] <= data;
      9'd415 : datareg[415*d+d-1:415*d] <= data;


      9'd416 : datareg[416*d+d-1:416*d] <= data;
      9'd417 : datareg[417*d+d-1:417*d] <= data;
      9'd418 : datareg[418*d+d-1:418*d] <= data;
      9'd419 : datareg[419*d+d-1:419*d] <= data;
      9'd420 : datareg[420*d+d-1:420*d] <= data;
      9'd421 : datareg[421*d+d-1:421*d] <= data;
      9'd422 : datareg[422*d+d-1:422*d] <= data;
      9'd423 : datareg[423*d+d-1:423*d] <= data;

      9'd424 : datareg[424*d+d-1:424*d] <= data;
      9'd425 : datareg[425*d+d-1:425*d] <= data;
      9'd426 : datareg[426*d+d-1:426*d] <= data;
      9'd427 : datareg[427*d+d-1:427*d] <= data;
      9'd428 : datareg[428*d+d-1:428*d] <= data;
      9'd429 : datareg[429*d+d-1:429*d] <= data;
      9'd430 : datareg[430*d+d-1:430*d] <= data;
      9'd431 : datareg[431*d+d-1:431*d] <= data;

      9'd432 : datareg[432*d+d-1:432*d] <= data;
      9'd433 : datareg[433*d+d-1:433*d] <= data;
      9'd434 : datareg[434*d+d-1:434*d] <= data;
      9'd435 : datareg[435*d+d-1:435*d] <= data;
      9'd436 : datareg[436*d+d-1:436*d] <= data;
      9'd437 : datareg[437*d+d-1:437*d] <= data;
      9'd438 : datareg[438*d+d-1:438*d] <= data;
      9'd439 : datareg[439*d+d-1:439*d] <= data;

      9'd440 : datareg[440*d+d-1:440*d] <= data;
      9'd441 : datareg[441*d+d-1:441*d] <= data;
      9'd442 : datareg[442*d+d-1:442*d] <= data;
      9'd443 : datareg[443*d+d-1:443*d] <= data;
      9'd444 : datareg[444*d+d-1:444*d] <= data;
      9'd445 : datareg[445*d+d-1:445*d] <= data;
      9'd446 : datareg[446*d+d-1:446*d] <= data;
      9'd447 : datareg[447*d+d-1:447*d] <= data;


      9'd448 : datareg[448*d+d-1:448*d] <= data;
      9'd449 : datareg[449*d+d-1:449*d] <= data;
      9'd450 : datareg[450*d+d-1:450*d] <= data;
      9'd451 : datareg[451*d+d-1:451*d] <= data;
      9'd452 : datareg[452*d+d-1:452*d] <= data;
      9'd453 : datareg[453*d+d-1:453*d] <= data;
      9'd454 : datareg[454*d+d-1:454*d] <= data;
      9'd455 : datareg[455*d+d-1:455*d] <= data;

      9'd456 : datareg[456*d+d-1:456*d] <= data;
      9'd457 : datareg[457*d+d-1:457*d] <= data;
      9'd458 : datareg[458*d+d-1:458*d] <= data;
      9'd459 : datareg[459*d+d-1:459*d] <= data;
      9'd460 : datareg[460*d+d-1:460*d] <= data;
      9'd461 : datareg[461*d+d-1:461*d] <= data;
      9'd462 : datareg[462*d+d-1:462*d] <= data;
      9'd463 : datareg[463*d+d-1:463*d] <= data;

      9'd464 : datareg[464*d+d-1:464*d] <= data;
      9'd465 : datareg[465*d+d-1:465*d] <= data;
      9'd466 : datareg[466*d+d-1:466*d] <= data;
      9'd467 : datareg[467*d+d-1:467*d] <= data;
      9'd468 : datareg[468*d+d-1:468*d] <= data;
      9'd469 : datareg[469*d+d-1:469*d] <= data;
      9'd470 : datareg[470*d+d-1:470*d] <= data;
      9'd471 : datareg[471*d+d-1:471*d] <= data;

      9'd472 : datareg[472*d+d-1:472*d] <= data;
      9'd473 : datareg[473*d+d-1:473*d] <= data;
      9'd474 : datareg[474*d+d-1:474*d] <= data;
      9'd475 : datareg[475*d+d-1:475*d] <= data;
      9'd476 : datareg[476*d+d-1:476*d] <= data;
      9'd477 : datareg[477*d+d-1:477*d] <= data;
      9'd478 : datareg[478*d+d-1:478*d] <= data;
      9'd479 : datareg[479*d+d-1:479*d] <= data;


      9'd480 : datareg[480*d+d-1:480*d] <= data;
      9'd481 : datareg[481*d+d-1:481*d] <= data;
      9'd482 : datareg[482*d+d-1:482*d] <= data;
      9'd483 : datareg[483*d+d-1:483*d] <= data;
      9'd484 : datareg[484*d+d-1:484*d] <= data;
      9'd485 : datareg[485*d+d-1:485*d] <= data;
      9'd486 : datareg[486*d+d-1:486*d] <= data;
      9'd487 : datareg[487*d+d-1:487*d] <= data;

      9'd488 : datareg[488*d+d-1:488*d] <= data;
      9'd489 : datareg[489*d+d-1:489*d] <= data;
      9'd490 : datareg[490*d+d-1:490*d] <= data;
      9'd491 : datareg[491*d+d-1:491*d] <= data;
      9'd492 : datareg[492*d+d-1:492*d] <= data;
      9'd493 : datareg[493*d+d-1:493*d] <= data;
      9'd494 : datareg[494*d+d-1:494*d] <= data;
      9'd495 : datareg[495*d+d-1:495*d] <= data;

      9'd496 : datareg[496*d+d-1:496*d] <= data;
      9'd497 : datareg[497*d+d-1:497*d] <= data;
      9'd498 : datareg[498*d+d-1:498*d] <= data;
      9'd499 : datareg[499*d+d-1:499*d] <= data;
      9'd500 : datareg[500*d+d-1:500*d] <= data;
      9'd501 : datareg[501*d+d-1:501*d] <= data;
      9'd502 : datareg[502*d+d-1:502*d] <= data;
      9'd503 : datareg[503*d+d-1:503*d] <= data;

      9'd504 : datareg[504*d+d-1:504*d] <= data;
      9'd505 : datareg[505*d+d-1:505*d] <= data;
      9'd506 : datareg[506*d+d-1:506*d] <= data;
      9'd507 : datareg[507*d+d-1:507*d] <= data;
      9'd508 : datareg[508*d+d-1:508*d] <= data;
      9'd509 : datareg[509*d+d-1:509*d] <= data;
      9'd510 : datareg[510*d+d-1:510*d] <= data;
      9'd511 : datareg[511*d+d-1:511*d] <= data;

     default   datareg <= 'b1;
     endcase
   end
 end


// - constantly output full reg contents

 assign out = datareg[wp-1:0];


// ===========================================

endmodule





// *******************************************
//   Deluge System Operation Control
//
// *******************************************

module  DLG_sControl (keyflag,done,datain,wr, pause,rekey, clk,gpause,aclr);
parameter d = 16;

// ===========================================

 input keyflag,done;
 input [d-1:0] datain;
 input  wr;

 output reg pause;
 output reg rekey;
 
 input  clk,gpause,aclr;

// ===========================================

// - pause if key found, or search finished
// - unpause if a '1' is writen to system

 always@(posedge clk, posedge aclr) begin
   if(aclr)           pause <= 'b0;
   else if(!wr)       pause <= |{keyflag,done,gpause};
   else if(datain[0]||datain[1]) pause <= 'b0; 
  end


// - rekey system if lsb writen to

 always@(posedge clk, posedge aclr) begin
   if(aclr)           rekey <= 'b0;
   else if(!wr)       rekey <= 'b0;
   else if(datain[0]) rekey <= 1'b1;
  end


// ===========================================

endmodule





// *******************************************
//   Deluge System Output Monitor
//
// *******************************************

module  DLG_sMonitor (keyflag,done, dataout, clk,aclr);
parameter d = 16;

// ===========================================

 input keyflag,done;

 output reg [d-1:0] dataout;
 
 input  clk,aclr;

// ===========================================

// - constantly output status
// -- keyfound = data[1]
// -- finished = data[0]

 always@(posedge clk, posedge aclr) begin
   if(aclr)  dataout <= 'b0;
   else      dataout <= { {(d-2){1'b0}},{keyflag,done} };
  end


// ===========================================

endmodule





// *******************************************
//   Deluge Ouput Register   (128bit max)
//
// *******************************************

module  DLG_oReg (data,addr,wr, out, clk,aclr);
parameter w = 80, d = 16;

// ===========================================

 input [w-1:0] data;
 input [3-1:0] addr;
 input  wr;

 output reg [d-1:0] out;

 input  clk,aclr;

 reg [128-1:0] datareg;


// ===========================================


// - constantly fill reg with wide data

 always@(posedge clk, posedge aclr) begin
   if(aclr)    datareg <= 'b0;
   else if(wr) datareg <= { {(128-w){1'b0}},data };
   else        datareg <= { {(128-w){1'b0}},data };  //useful for monitoring
  end


// - output reg contents in parts

 always@(posedge clk, posedge aclr) begin
   if(aclr)  out <= 'b0;
   else case(addr)
      3'h0 : out <= datareg[0*d+d-1:0*d];
      3'h1 : out <= datareg[1*d+d-1:1*d];
      3'h2 : out <= datareg[2*d+d-1:2*d];
      3'h3 : out <= datareg[3*d+d-1:3*d];
      3'h4 : out <= datareg[4*d+d-1:4*d];
      3'h5 : out <= datareg[5*d+d-1:5*d];
      3'h6 : out <= datareg[6*d+d-1:6*d];
      3'h7 : out <= datareg[7*d+d-1:7*d];
     default  out <= 'b0;
     endcase
  end


// ===========================================

endmodule





// *******************************************
//   Deluge Data Ouput Multiplexer
//
// *******************************************

module  DLG_OutMUX (data0,data1,select, out);
parameter d = 16;

// ===========================================

 input [d-1:0] data0,data1;
 input  select;

 output [d-1:0] out;

// ===========================================


 assign out = select ?  data1: data0;


// ===========================================

endmodule





// ------------------------------------------------------
// ------------------------------------------------------




 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 // SYSTEM OPERATION INFO
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 //  xclk    |  interface clock - optimise for compatability       |
 //  clk     |  core clock      - optimise for max search speed    |
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 //  xaclr   |  reset system interface to default state            |
 //          |          (clears plain, key & IV registers)         |
 //  aclr    |  reset search core to default state                 |
 //          |          (use after loading new register key)       |
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
 //  gpause  |  pause core (sleep)                                 |
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 //  cs, wIr | addr[msb..lsb]  |     data      | NOTES...          |
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 //   0   X  |        X        |     ZZZZ      |                   |
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 //   1   0  |     X     I  0  | {0001} =>done | read status reg   |
 //          |           I     | {0002} =>flag |  (outputing data) |
 //          |  X  I  C  I  1  | {search key}  | read o/p key reg  |
 //          |     I     I     |               |  (C selects part) |
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 //   1   1  |  00 I     X     |               | write ack         |
 //          |     I           |     XXX1      |     (rekeys core) |
 //          |     I           |     XXX2      |     (clears flag) |
 //          |  01 I  X  I  A  |     rKey      | write reg key     |
 //          |     I     I     |               |       (A -> part) |
 //          |  10 I  X  I  A  |      IV       | write IV          |
 //          |     I     I     |               |       (A -> part) |
 //          |  11 I  J  I  C  |     plain     | write plainstream |
 //          |     I     I     |               |   (J -> position) |
 //          |     I     I     |               |   (C -> part    ) |
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 // **   raised flag indicates a successful key has been found
 // **   note: a 2 cylce latency is present on outputs
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -




// ------------------------------------------------------
// ------------------------------------------------------






// *****************************************
//   Deluge - Search Core
//
// *****************************************

module Del_Core (rKey,IV,set,plainMem,datain,adr,wr, oKey,keyflag,done, clk,pause,aclr, xclk);
parameter  i = 2, j = 4, v = 31, h = 0, w = 16, k = 80,
           n = 2**i,  //number of ciphers
           p = 2**j,  //number of plainstreams
           s = 2**h,  //compaction factor
           f = p*n,   //flags before compaction
           fs = f/s;  //flags after compaction

`ifdef DEL_NO_ID  parameter nid = 1; `else parameter nid = 0; `endif
`ifdef DEL_NO_PID  parameter npid = 1; `else parameter npid = 0; `endif

// =========================================

 input [k-v-i-1:0] rKey;
 input [32-1:0] IV;
 input set;
 input [w*p-1:0] plainMem;

 input [w-1:0] datain;
 input [j+2:0] adr;
 input wr;

 output [k-1:0] oKey;
 output keyflag;
 output reg done;

 input xclk, clk,pause,aclr;

 wire [fs-1:0] fd;
 wire [v-1:0] vd;
 wire fdone;
 wire empty,read;

 wire [j+npid-h-1:0] pid;
 wire [i+nid-1:0] id;
 wire [v-1:0] vKey;
 wire em,rd;

 reg [5:0] count;
 
 // - - - - - - - - - - - - - - - - - - - - - -
 // for system debug add
 // - - - - - - - - - - - - - - - - - - - - - -
 // * module head: ,empty,em,plainstream_,plainstream1_,pid_);
 // - - - - - - - - - - - - - - - - - - - - - -
 // * output empty,em; 
 // * output [w-1:0] plainstream_,plainstream1_;
 // output [j-h-1:0] pid_;
 // - - - - - - - - - - - - - - - - - - - - - -
 // + module DEL_BACKEND_SEARCH: , plainstream_,plainstream1_);
 // assign pid_ = pid;
 // - - - - - - - - - - - - - - - - - - - - - -

 // - - - - - - - - - - - - - - - - - - - - - -
 // for hardware test add
 // - - - - - - - - - - - - - - - - - - - - - -
 // * module head: ,pid_);
 // - - - - - - - - - - - - - - - - - - - - - -
 // output [j-h-1:0] pid_;
 // assign pid_ = pid;
 // - - - - - - - - - - - - - - - - - - - - - -


// =========================================


// *** SEARCH CORE ***

 `DEL_FRONTEND_SEARCH #(.i(i),.j(j),.v(v),.h(h),.k(k)) DLF (rKey,IV,set,plainMem, fd,vd,fdone,empty, read, clk,pause,aclr);

  Del_TranslationUnit #(.i(i),.j(j-h),.v(v),.f(fs))    DLT (fd,vd,empty, pid,id,vKey,read,em, rd, clk,pause,aclr);

 `DEL_BACKEND_SEARCH  #(.i(i),.j(j),.v(v),.h(h),.k(k))  DLB (pid,id,vKey,rKey,IV,datain,adr,wr,em, rd,oKey,keyflag, clk,pause,aclr, xclk);



// *** DONE CONTROL ***

 // - allow system to settle for a few cycles

 always @(posedge clk, posedge aclr) begin
   if(aclr)        count <= 1'b0;
   else if(set)    count <= 1'b0;
   else if(!pause && (fdone && count<63)) count <= count + 1'b1;  //only count 40 actually needed
  end
  
 // - before sending done signal to output
 
 always @(posedge clk, posedge aclr) begin
   if(aclr)        done <= 1'b0;
   else if(!pause) done <= &{fdone,empty,em,{&count}};
  end


// =========================================

endmodule


