
// *****************************************
//   Trivium - Stream Cipher
//
//   ver:eStream
//   status: ok   Mar/06
//
// *****************************************
//   Verilog Code
//   Iain Devlin 2006
//
// *****************************************






// *****************************************
//   Trivium 1
//
// *****************************************

module Trivium_1 (Key,IV,rekey,pause, keystream,valid, clk,aclr);
parameter w = 1;

// =========================================

 input  [80-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w-1:0] keystream;
 output valid;

 input  clk,aclr;

 wire [93-1:0]  f;
 wire [84-1:0]  g;
 wire [111-1:0] h;

 wire [w-1:0]  fx,gx,hx;

 wire [w-1:0] z;


// =========================================

 TriviumReg #(.w(w)) REG0 (Key,{IV,48'h0},rekey,pause, fx,gx,hx, f,g,h, clk,aclr);

 TriviumControl #(.w(w),.j(11)) CTRL0 (rekey,pause,z, keystream,valid, clk,aclr);

 Tri_Gen #(.i(0)) TG0 (f,g,h, fx[w-1-0],gx[w-1-0],hx[w-1-0],z[0]);


// =========================================

endmodule






// *****************************************
//   Trivium - Register Setup
//
// *****************************************

module TriviumReg (Key,IV,rekey,pause, fx,gx,hx, f,g,h, clk,aclr);
parameter w = 1;

// =========================================

 input  [80-1:0]  Key;
 input  [80-1:0]  IV;
 input  rekey,pause;

 input [w-1:0]  fx,gx,hx;

 output reg [93-1:0]  f;
 output reg [84-1:0]  g;
 output reg [111-1:0] h;

 input  clk,aclr;


// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)        f <= 93'b0;
   else if(rekey)  f <= {13'b0,Key};
   else if(!pause) f <= {f[93-1-w:0],fx};
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        g <= 84'b0;
   else if(rekey)  g <= {4'b0,IV};
   else if(!pause) g <= {g[84-1-w:0],gx};
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        h <= 111'b0;
   else if(rekey)  h <= {{(3){1'b1}},108'b0};
   else if(!pause) h <= {h[111-1-w:0],hx};
  end


// =========================================

endmodule





// *****************************************
//   Trivium - Control Unit
//
// *****************************************

module TriviumControl (set,pause,z, keystream,valid, clk,aclr);
parameter w = 1;
parameter i = (4*288)/w, j = 11;

// =========================================

 input  set,pause;

 output reg valid;

 input  clk,aclr;

 reg   [j-1:0] count;
 reg   enable;

 input [w-1:0] z;
 output reg [w-1:0] keystream;


// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)        count <= {(j){1'b0}};
   else if(set)    count <= {(j){1'b0}};
   else if(!pause) count <= count + 1'b1;
  end

 always@(posedge clk, posedge aclr) begin
   if(aclr)        enable <= 1'b0;
   else if(set)    enable <= 1'b0;
   else if(!pause)
             if(count==(i-1) || enable)  enable <= 1'b1;
             else  enable <= 1'b0;
  end

 always@(posedge clk, posedge aclr) begin
   if(aclr)        valid <= 1'b0;
   else if(set)    valid <= 1'b0;
   else if(!pause) valid <= enable;
  end

 always@(posedge clk, posedge aclr) begin
   if(aclr)         keystream <= {(w){1'b0}};
   else if(!enable) keystream <= {(w){1'b0}};
   else if(!pause)  keystream <= z;
  end


// =========================================

endmodule





// ------------------------------------------------------
// ------------------------------------------------------






// *****************************************
//   Trivium - Algorithm Core
//
// *****************************************

module Tri_Gen (f,g,h, fx,gx,hx,z);
parameter i=0;

// =========================================

input  [92:0]  f;
input  [83:0]  g;
input  [110:0] h;

output fx,gx,hx;
output z;

wire t1,t2,t3;

// =========================================


 Tri_func F0 (h[65-i],h[110-i],h[109-i],h[108-i],f[68-i], t3,fx);
 Tri_func G0 (f[65-i],f[92-i],f[91-i],f[90-i],g[77-i],    t1,gx);
 Tri_func H0 (g[68-i],g[83-i],g[82-i],g[81-i],h[86-i],    t2,hx);

 assign z = t1^t2^t3;


// =========================================

endmodule




module Tri_func (s243,s288,s287,s286,s69, t,s);

// =========================================

input  s243,s288,s286,s287,s69;
output reg t,s;

// =========================================

 always@* begin
   t <= s243 ^ s288;
   s <= t ^ s287&s286 ^ s69;  
  end

// =========================================

endmodule





// ------------------------------------------------------
// ------------------------------------------------------






// *****************************************
//   Trivium 2
//
// *****************************************

module Trivium_2 (Key,IV,rekey,pause, keystream,valid, clk,aclr);
parameter w = 2;

// =========================================

 input  [80-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w-1:0] keystream;
 output valid;

 input  clk,aclr;

 wire [93-1:0]  f;
 wire [84-1:0]  g;
 wire [111-1:0] h;

 wire [w-1:0]  fx,gx,hx;

 wire [w-1:0] z;


// =========================================

 TriviumReg #(.w(w)) REG0 (Key,{IV,48'h0},rekey,pause, fx,gx,hx, f,g,h, clk,aclr);

 TriviumControl  #(.w(w),.j(10)) CTRL0 (rekey,pause,z, keystream,valid, clk,aclr);

 Tri_Gen #(.i(0)) TG0 (f,g,h, fx[w-1-0],gx[w-1-0],hx[w-1-0],z[0]);
 Tri_Gen #(.i(1)) TG1 (f,g,h, fx[w-1-1],gx[w-1-1],hx[w-1-1],z[1]);


// =========================================

endmodule






// *****************************************
//   Trivium 8
//
// *****************************************

module Trivium_8 (Key,IV,rekey,pause, keystream,valid, clk,aclr);
parameter w = 8;

// =========================================

 input  [80-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w-1:0] keystream;
 output valid;

 input  clk,aclr;

 wire [93-1:0]  f;
 wire [84-1:0]  g;
 wire [111-1:0] h;

 wire [w-1:0]  fx,gx,hx;

 wire [w-1:0] z;


// =========================================

 TriviumReg #(.w(w)) REG0 (Key,{IV,48'h0},rekey,pause, fx,gx,hx, f,g,h, clk,aclr);

 TriviumControl  #(.w(w),.j(8)) CTRL0 (rekey,pause,z, keystream,valid, clk,aclr);

 Tri_Gen #(.i(0))  TG00 (f,g,h, fx[w-1-0], gx[w-1-0], hx[w-1-0], z[0] );
 Tri_Gen #(.i(1))  TG01 (f,g,h, fx[w-1-1], gx[w-1-1], hx[w-1-1], z[1] );
 Tri_Gen #(.i(2))  TG02 (f,g,h, fx[w-1-2], gx[w-1-2], hx[w-1-2], z[2] );
 Tri_Gen #(.i(3))  TG03 (f,g,h, fx[w-1-3], gx[w-1-3], hx[w-1-3], z[3] );
 Tri_Gen #(.i(4))  TG04 (f,g,h, fx[w-1-4], gx[w-1-4], hx[w-1-4], z[4] );
 Tri_Gen #(.i(5))  TG05 (f,g,h, fx[w-1-5], gx[w-1-5], hx[w-1-5], z[5] );
 Tri_Gen #(.i(6))  TG06 (f,g,h, fx[w-1-6], gx[w-1-6], hx[w-1-6], z[6] );
 Tri_Gen #(.i(7))  TG07 (f,g,h, fx[w-1-7], gx[w-1-7], hx[w-1-7], z[7] );


// =========================================

endmodule





// *****************************************
//   Trivium 16
//
// *****************************************

module Trivium_16 (Key,IV,rekey,pause, keystream,valid, clk,aclr);
parameter w = 16;

// =========================================

 input  [80-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w-1:0] keystream;
 output valid;

 input  clk,aclr;

 wire [93-1:0]  f;
 wire [84-1:0]  g;
 wire [111-1:0] h;

 wire [w-1:0]  fx,gx,hx;

 wire [w-1:0] z;


// =========================================

 TriviumReg #(.w(w)) REG0 (Key,{IV,48'h0},rekey,pause, fx,gx,hx, f,g,h, clk,aclr);

 TriviumControl  #(.w(w),.j(7)) CTRL0 (rekey,pause,z, keystream,valid, clk,aclr);

 Tri_Gen #(.i(0))  TG00 (f,g,h, fx[w-1-0], gx[w-1-0], hx[w-1-0], z[0] );
 Tri_Gen #(.i(1))  TG01 (f,g,h, fx[w-1-1], gx[w-1-1], hx[w-1-1], z[1] );
 Tri_Gen #(.i(2))  TG02 (f,g,h, fx[w-1-2], gx[w-1-2], hx[w-1-2], z[2] );
 Tri_Gen #(.i(3))  TG03 (f,g,h, fx[w-1-3], gx[w-1-3], hx[w-1-3], z[3] );
 Tri_Gen #(.i(4))  TG04 (f,g,h, fx[w-1-4], gx[w-1-4], hx[w-1-4], z[4] );
 Tri_Gen #(.i(5))  TG05 (f,g,h, fx[w-1-5], gx[w-1-5], hx[w-1-5], z[5] );
 Tri_Gen #(.i(6))  TG06 (f,g,h, fx[w-1-6], gx[w-1-6], hx[w-1-6], z[6] );
 Tri_Gen #(.i(7))  TG07 (f,g,h, fx[w-1-7], gx[w-1-7], hx[w-1-7], z[7] );
 Tri_Gen #(.i(8))  TG08 (f,g,h, fx[w-1-8], gx[w-1-8], hx[w-1-8], z[8] );
 Tri_Gen #(.i(9))  TG09 (f,g,h, fx[w-1-9], gx[w-1-9], hx[w-1-9], z[9] );
 Tri_Gen #(.i(10)) TG10 (f,g,h, fx[w-1-10],gx[w-1-10],hx[w-1-10],z[10]);
 Tri_Gen #(.i(11)) TG11 (f,g,h, fx[w-1-11],gx[w-1-11],hx[w-1-11],z[11]);
 Tri_Gen #(.i(12)) TG12 (f,g,h, fx[w-1-12],gx[w-1-12],hx[w-1-12],z[12]);
 Tri_Gen #(.i(13)) TG13 (f,g,h, fx[w-1-13],gx[w-1-13],hx[w-1-13],z[13]);
 Tri_Gen #(.i(14)) TG14 (f,g,h, fx[w-1-14],gx[w-1-14],hx[w-1-14],z[14]);
 Tri_Gen #(.i(15)) TG15 (f,g,h, fx[w-1-15],gx[w-1-15],hx[w-1-15],z[15]);


// =========================================

endmodule






// *****************************************
//   Trivium 64
//
// *****************************************

module Trivium_64 (Key,IV,rekey,pause, keystream,valid, clk,aclr);
parameter w = 64;

// =========================================

 input  [80-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w-1:0] keystream;
 output valid;

 input  clk,aclr;

 wire [93-1:0]  f;
 wire [84-1:0]  g;
 wire [111-1:0] h;

 wire [w-1:0]  fx,gx,hx;

 wire [w-1:0] z;


// =========================================

 TriviumReg #(.w(w)) REG0 (Key,{IV,48'h0},rekey,pause, fx,gx,hx, f,g,h, clk,aclr);

 TriviumControl  #(.w(w),.j(5)) CTRL0 (rekey,pause,z, keystream,valid, clk,aclr);

 Tri_Gen #(.i(0))  TG00 (f,g,h, fx[w-1-0], gx[w-1-0], hx[w-1-0], z[0] );
 Tri_Gen #(.i(1))  TG01 (f,g,h, fx[w-1-1], gx[w-1-1], hx[w-1-1], z[1] );
 Tri_Gen #(.i(2))  TG02 (f,g,h, fx[w-1-2], gx[w-1-2], hx[w-1-2], z[2] );
 Tri_Gen #(.i(3))  TG03 (f,g,h, fx[w-1-3], gx[w-1-3], hx[w-1-3], z[3] );
 Tri_Gen #(.i(4))  TG04 (f,g,h, fx[w-1-4], gx[w-1-4], hx[w-1-4], z[4] );
 Tri_Gen #(.i(5))  TG05 (f,g,h, fx[w-1-5], gx[w-1-5], hx[w-1-5], z[5] );
 Tri_Gen #(.i(6))  TG06 (f,g,h, fx[w-1-6], gx[w-1-6], hx[w-1-6], z[6] );
 Tri_Gen #(.i(7))  TG07 (f,g,h, fx[w-1-7], gx[w-1-7], hx[w-1-7], z[7] );
 Tri_Gen #(.i(8))  TG08 (f,g,h, fx[w-1-8], gx[w-1-8], hx[w-1-8], z[8] );
 Tri_Gen #(.i(9))  TG09 (f,g,h, fx[w-1-9], gx[w-1-9], hx[w-1-9], z[9] );
 Tri_Gen #(.i(10)) TG10 (f,g,h, fx[w-1-10],gx[w-1-10],hx[w-1-10],z[10]);
 Tri_Gen #(.i(11)) TG11 (f,g,h, fx[w-1-11],gx[w-1-11],hx[w-1-11],z[11]);
 Tri_Gen #(.i(12)) TG12 (f,g,h, fx[w-1-12],gx[w-1-12],hx[w-1-12],z[12]);
 Tri_Gen #(.i(13)) TG13 (f,g,h, fx[w-1-13],gx[w-1-13],hx[w-1-13],z[13]);
 Tri_Gen #(.i(14)) TG14 (f,g,h, fx[w-1-14],gx[w-1-14],hx[w-1-14],z[14]);
 Tri_Gen #(.i(15)) TG15 (f,g,h, fx[w-1-15],gx[w-1-15],hx[w-1-15],z[15]);

 Tri_Gen #(.i(16)) TG16 (f,g,h, fx[w-1-16],gx[w-1-16],hx[w-1-16],z[16]);
 Tri_Gen #(.i(17)) TG17 (f,g,h, fx[w-1-17],gx[w-1-17],hx[w-1-17],z[17]);
 Tri_Gen #(.i(18)) TG18 (f,g,h, fx[w-1-18],gx[w-1-18],hx[w-1-18],z[18]);
 Tri_Gen #(.i(19)) TG19 (f,g,h, fx[w-1-19],gx[w-1-19],hx[w-1-19],z[19]);
 Tri_Gen #(.i(20)) TG20 (f,g,h, fx[w-1-20],gx[w-1-20],hx[w-1-20],z[20]);
 Tri_Gen #(.i(21)) TG21 (f,g,h, fx[w-1-21],gx[w-1-21],hx[w-1-21],z[21]);
 Tri_Gen #(.i(22)) TG22 (f,g,h, fx[w-1-22],gx[w-1-22],hx[w-1-22],z[22]);
 Tri_Gen #(.i(23)) TG23 (f,g,h, fx[w-1-23],gx[w-1-23],hx[w-1-23],z[23]);
 Tri_Gen #(.i(24)) TG24 (f,g,h, fx[w-1-24],gx[w-1-24],hx[w-1-24],z[24]);
 Tri_Gen #(.i(25)) TG25 (f,g,h, fx[w-1-25],gx[w-1-25],hx[w-1-25],z[25]);
 Tri_Gen #(.i(26)) TG26 (f,g,h, fx[w-1-26],gx[w-1-26],hx[w-1-26],z[26]);
 Tri_Gen #(.i(27)) TG27 (f,g,h, fx[w-1-27],gx[w-1-27],hx[w-1-27],z[27]);
 Tri_Gen #(.i(28)) TG28 (f,g,h, fx[w-1-28],gx[w-1-28],hx[w-1-28],z[28]);
 Tri_Gen #(.i(29)) TG29 (f,g,h, fx[w-1-29],gx[w-1-29],hx[w-1-29],z[29]);
 Tri_Gen #(.i(30)) TG30 (f,g,h, fx[w-1-30],gx[w-1-30],hx[w-1-30],z[30]);
 Tri_Gen #(.i(31)) TG31 (f,g,h, fx[w-1-31],gx[w-1-31],hx[w-1-31],z[31]);

 Tri_Gen #(.i(32)) TG32 (f,g,h, fx[w-1-32],gx[w-1-32],hx[w-1-32],z[32]);
 Tri_Gen #(.i(33)) TG33 (f,g,h, fx[w-1-33],gx[w-1-33],hx[w-1-33],z[33]);
 Tri_Gen #(.i(34)) TG34 (f,g,h, fx[w-1-34],gx[w-1-34],hx[w-1-34],z[34]);
 Tri_Gen #(.i(35)) TG35 (f,g,h, fx[w-1-35],gx[w-1-35],hx[w-1-35],z[35]);
 Tri_Gen #(.i(36)) TG36 (f,g,h, fx[w-1-36],gx[w-1-36],hx[w-1-36],z[36]);
 Tri_Gen #(.i(37)) TG37 (f,g,h, fx[w-1-37],gx[w-1-37],hx[w-1-37],z[37]);
 Tri_Gen #(.i(38)) TG38 (f,g,h, fx[w-1-38],gx[w-1-38],hx[w-1-38],z[38]);
 Tri_Gen #(.i(39)) TG39 (f,g,h, fx[w-1-39],gx[w-1-39],hx[w-1-39],z[39]);
 Tri_Gen #(.i(40)) TG40 (f,g,h, fx[w-1-40],gx[w-1-40],hx[w-1-40],z[40]);
 Tri_Gen #(.i(41)) TG41 (f,g,h, fx[w-1-41],gx[w-1-41],hx[w-1-41],z[41]);
 Tri_Gen #(.i(42)) TG42 (f,g,h, fx[w-1-42],gx[w-1-42],hx[w-1-42],z[42]);
 Tri_Gen #(.i(43)) TG43 (f,g,h, fx[w-1-43],gx[w-1-43],hx[w-1-43],z[43]);
 Tri_Gen #(.i(44)) TG44 (f,g,h, fx[w-1-44],gx[w-1-44],hx[w-1-44],z[44]);
 Tri_Gen #(.i(45)) TG45 (f,g,h, fx[w-1-45],gx[w-1-45],hx[w-1-45],z[45]);
 Tri_Gen #(.i(46)) TG46 (f,g,h, fx[w-1-46],gx[w-1-46],hx[w-1-46],z[46]);
 Tri_Gen #(.i(47)) TG47 (f,g,h, fx[w-1-47],gx[w-1-47],hx[w-1-47],z[47]);

 Tri_Gen #(.i(48)) TG48 (f,g,h, fx[w-1-48],gx[w-1-48],hx[w-1-48],z[48]);
 Tri_Gen #(.i(49)) TG49 (f,g,h, fx[w-1-49],gx[w-1-49],hx[w-1-49],z[49]);
 Tri_Gen #(.i(50)) TG50 (f,g,h, fx[w-1-50],gx[w-1-50],hx[w-1-50],z[50]);
 Tri_Gen #(.i(51)) TG51 (f,g,h, fx[w-1-51],gx[w-1-51],hx[w-1-51],z[51]);
 Tri_Gen #(.i(52)) TG52 (f,g,h, fx[w-1-52],gx[w-1-52],hx[w-1-52],z[52]);
 Tri_Gen #(.i(53)) TG53 (f,g,h, fx[w-1-53],gx[w-1-53],hx[w-1-53],z[53]);
 Tri_Gen #(.i(54)) TG54 (f,g,h, fx[w-1-54],gx[w-1-54],hx[w-1-54],z[54]);
 Tri_Gen #(.i(55)) TG55 (f,g,h, fx[w-1-55],gx[w-1-55],hx[w-1-55],z[55]);
 Tri_Gen #(.i(56)) TG56 (f,g,h, fx[w-1-56],gx[w-1-56],hx[w-1-56],z[56]);
 Tri_Gen #(.i(57)) TG57 (f,g,h, fx[w-1-57],gx[w-1-57],hx[w-1-57],z[57]);
 Tri_Gen #(.i(58)) TG58 (f,g,h, fx[w-1-58],gx[w-1-58],hx[w-1-58],z[58]);
 Tri_Gen #(.i(59)) TG59 (f,g,h, fx[w-1-59],gx[w-1-59],hx[w-1-59],z[59]);
 Tri_Gen #(.i(60)) TG60 (f,g,h, fx[w-1-60],gx[w-1-60],hx[w-1-60],z[60]);
 Tri_Gen #(.i(61)) TG61 (f,g,h, fx[w-1-61],gx[w-1-61],hx[w-1-61],z[61]);
 Tri_Gen #(.i(62)) TG62 (f,g,h, fx[w-1-62],gx[w-1-62],hx[w-1-62],z[62]);
 Tri_Gen #(.i(63)) TG63 (f,g,h, fx[w-1-63],gx[w-1-63],hx[w-1-63],z[63]);



// =========================================

endmodule





// ------------------------------------------------------
// ------------------------------------------------------






// *****************************************
//   Trivium 64 - Pipelined for keysearch
//
// *****************************************

module TriviumP64 (Key,IV,rekey,pause, keystream,valid, clk,aclr);
parameter w = 64;

// =========================================

 input  [80-1:0]  Key;
 input  [32-1:0]  IV;
 input  rekey,pause;

 output [w-1:0] keystream;
 output valid;

 input  clk,aclr;

 wire [93-1:0]  f0,f1,f2,f3,f4,f5,f6,f7,f8,f9,f10,f11,f12,f13,f14,f15,f16,f17,f18;
 wire [84-1:0]  g0,g1,g2,g3,g4,g5,g6,g7,g8,g9,g10,g11,g12,g13,g14,g15,g16,g17,g18; 
 wire [111-1:0] h0,h1,h2,h3,h4,h5,h6,h7,h8,h9,h10,h11,h12,h13,h14,h15,h16,h17,h18; 
 wire  val0,val1,val2,val3,val4,val5,val6,val7,val8,val9,
       val10,val11,val12,val13,val14,val15,val16,val17,val18;


// =========================================

 TriviumP_Start #(.w(w)) S0  ({13'b0,Key},{4'b0,IV,48'b0},{{(3){1'b1}},108'b0},rekey,pause, f0,g0,h0,val0, clk,aclr);
 TriviumP_Init  #(.w(w)) I1  (f0,g0,h0,val0,pause, f1,g1,h1,val1, clk,aclr);
 TriviumP_Init  #(.w(w)) I2  (f1,g1,h1,val1,pause, f2,g2,h2,val2, clk,aclr);
 TriviumP_Init  #(.w(w)) I3  (f2,g2,h2,val2,pause, f3,g3,h3,val3, clk,aclr);
 TriviumP_Init  #(.w(w)) I4  (f3,g3,h3,val3,pause, f4,g4,h4,val4, clk,aclr);
 TriviumP_Init  #(.w(w)) I5  (f4,g4,h4,val4,pause, f5,g5,h5,val5, clk,aclr);
 TriviumP_Init  #(.w(w)) I6  (f5,g5,h5,val5,pause, f6,g6,h6,val6, clk,aclr);
 TriviumP_Init  #(.w(w)) I7  (f6,g6,h6,val6,pause, f7,g7,h7,val7, clk,aclr);
 TriviumP_Init  #(.w(w)) I8  (f7,g7,h7,val7,pause, f8,g8,h8,val8, clk,aclr);
 TriviumP_Init  #(.w(w)) I9  (f8,g8,h8,val8,pause, f9,g9,h9,val9, clk,aclr);
 TriviumP_Init  #(.w(w)) I10 (f9,g9,h9,val9,pause, f10,g10,h10,val10, clk,aclr);
 TriviumP_Init  #(.w(w)) I11 (f10,g10,h10,val10,pause, f11,g11,h11,val11, clk,aclr);
 TriviumP_Init  #(.w(w)) I12 (f11,g11,h11,val11,pause, f12,g12,h12,val12, clk,aclr);
 TriviumP_Init  #(.w(w)) I13 (f12,g12,h12,val12,pause, f13,g13,h13,val13, clk,aclr);
 TriviumP_Init  #(.w(w)) I14 (f13,g13,h13,val13,pause, f14,g14,h14,val14, clk,aclr);
 TriviumP_Init  #(.w(w)) I15 (f14,g14,h14,val14,pause, f15,g15,h15,val15, clk,aclr);
 TriviumP_Init  #(.w(w)) I16 (f15,g15,h15,val15,pause, f16,g16,h16,val16, clk,aclr);
 TriviumP_Init  #(.w(w)) I17 (f16,g16,h16,val16,pause, f17,g17,h17,val17, clk,aclr);
 TriviumP_Init  #(.w(w)) I18 (f17,g17,h17,val17,pause, f18,g18,h18,val18, clk,aclr);
 TriviumP_Out   #(.w(w)) O19 (f18,g18,h18,val18,pause, keystream,valid, clk,aclr);


// =========================================

endmodule






// *****************************************
//   Trivium - Pipeline Start Stage
//
// *****************************************

module TriviumP_Start (load_f,load_g,load_h,rekey,pause, f,g,h,valid, clk,aclr);
parameter w = 1;

// =========================================

 input [93-1:0]  load_f;
 input [84-1:0]  load_g;
 input [111-1:0] load_h;
 input  rekey,pause;

 output [93-1:0]  f;
 output [84-1:0]  g;
 output [111-1:0] h;
 output valid;

 input  clk,aclr;


// =========================================

 TriviumP_Reg #(.w(w)) REG0 (load_f,load_g,load_h,rekey,pause, f,g,h, clk,aclr);

 TriviumP_Control #(.w(w)) CTRL0 (rekey,pause,{(w){1'b0}}, ,valid, clk,aclr);


// =========================================

endmodule




// *****************************************
//   Trivium - Pipelined Initialise Stage
//
// *****************************************

module TriviumP_Init (f,g,h,rekey,pause, fReg,gReg,hReg,valid, clk,aclr);
parameter w = 1;

// =========================================

 input [93-1:0]  f;
 input [84-1:0]  g;
 input [111-1:0] h;
 input  rekey,pause;

 output [93-1:0]  fReg;
 output [84-1:0]  gReg;
 output [111-1:0] hReg;
 output valid;

 input  clk,aclr;

 wire [w-1:0]  fx,gx,hx;


// =========================================

 Tri_Gen #(.i(0))  TG00 (f,g,h, fx[w-1-0], gx[w-1-0], hx[w-1-0],      );
 Tri_Gen #(.i(1))  TG01 (f,g,h, fx[w-1-1], gx[w-1-1], hx[w-1-1],      );
 Tri_Gen #(.i(2))  TG02 (f,g,h, fx[w-1-2], gx[w-1-2], hx[w-1-2],      );
 Tri_Gen #(.i(3))  TG03 (f,g,h, fx[w-1-3], gx[w-1-3], hx[w-1-3],      );
 Tri_Gen #(.i(4))  TG04 (f,g,h, fx[w-1-4], gx[w-1-4], hx[w-1-4],      );
 Tri_Gen #(.i(5))  TG05 (f,g,h, fx[w-1-5], gx[w-1-5], hx[w-1-5],      );
 Tri_Gen #(.i(6))  TG06 (f,g,h, fx[w-1-6], gx[w-1-6], hx[w-1-6],      );
 Tri_Gen #(.i(7))  TG07 (f,g,h, fx[w-1-7], gx[w-1-7], hx[w-1-7],      );
 Tri_Gen #(.i(8))  TG08 (f,g,h, fx[w-1-8], gx[w-1-8], hx[w-1-8],      );
 Tri_Gen #(.i(9))  TG09 (f,g,h, fx[w-1-9], gx[w-1-9], hx[w-1-9],      );
 Tri_Gen #(.i(10)) TG10 (f,g,h, fx[w-1-10],gx[w-1-10],hx[w-1-10],     );
 Tri_Gen #(.i(11)) TG11 (f,g,h, fx[w-1-11],gx[w-1-11],hx[w-1-11],     );
 Tri_Gen #(.i(12)) TG12 (f,g,h, fx[w-1-12],gx[w-1-12],hx[w-1-12],     );
 Tri_Gen #(.i(13)) TG13 (f,g,h, fx[w-1-13],gx[w-1-13],hx[w-1-13],     );
 Tri_Gen #(.i(14)) TG14 (f,g,h, fx[w-1-14],gx[w-1-14],hx[w-1-14],     );
 Tri_Gen #(.i(15)) TG15 (f,g,h, fx[w-1-15],gx[w-1-15],hx[w-1-15],     );

 Tri_Gen #(.i(16)) TG16 (f,g,h, fx[w-1-16],gx[w-1-16],hx[w-1-16],     );
 Tri_Gen #(.i(17)) TG17 (f,g,h, fx[w-1-17],gx[w-1-17],hx[w-1-17],     );
 Tri_Gen #(.i(18)) TG18 (f,g,h, fx[w-1-18],gx[w-1-18],hx[w-1-18],     );
 Tri_Gen #(.i(19)) TG19 (f,g,h, fx[w-1-19],gx[w-1-19],hx[w-1-19],     );
 Tri_Gen #(.i(20)) TG20 (f,g,h, fx[w-1-20],gx[w-1-20],hx[w-1-20],     );
 Tri_Gen #(.i(21)) TG21 (f,g,h, fx[w-1-21],gx[w-1-21],hx[w-1-21],     );
 Tri_Gen #(.i(22)) TG22 (f,g,h, fx[w-1-22],gx[w-1-22],hx[w-1-22],     );
 Tri_Gen #(.i(23)) TG23 (f,g,h, fx[w-1-23],gx[w-1-23],hx[w-1-23],     );
 Tri_Gen #(.i(24)) TG24 (f,g,h, fx[w-1-24],gx[w-1-24],hx[w-1-24],     );
 Tri_Gen #(.i(25)) TG25 (f,g,h, fx[w-1-25],gx[w-1-25],hx[w-1-25],     );
 Tri_Gen #(.i(26)) TG26 (f,g,h, fx[w-1-26],gx[w-1-26],hx[w-1-26],     );
 Tri_Gen #(.i(27)) TG27 (f,g,h, fx[w-1-27],gx[w-1-27],hx[w-1-27],     );
 Tri_Gen #(.i(28)) TG28 (f,g,h, fx[w-1-28],gx[w-1-28],hx[w-1-28],     );
 Tri_Gen #(.i(29)) TG29 (f,g,h, fx[w-1-29],gx[w-1-29],hx[w-1-29],     );
 Tri_Gen #(.i(30)) TG30 (f,g,h, fx[w-1-30],gx[w-1-30],hx[w-1-30],     );
 Tri_Gen #(.i(31)) TG31 (f,g,h, fx[w-1-31],gx[w-1-31],hx[w-1-31],     );

 Tri_Gen #(.i(32)) TG32 (f,g,h, fx[w-1-32],gx[w-1-32],hx[w-1-32],     );
 Tri_Gen #(.i(33)) TG33 (f,g,h, fx[w-1-33],gx[w-1-33],hx[w-1-33],     );
 Tri_Gen #(.i(34)) TG34 (f,g,h, fx[w-1-34],gx[w-1-34],hx[w-1-34],     );
 Tri_Gen #(.i(35)) TG35 (f,g,h, fx[w-1-35],gx[w-1-35],hx[w-1-35],     );
 Tri_Gen #(.i(36)) TG36 (f,g,h, fx[w-1-36],gx[w-1-36],hx[w-1-36],     );
 Tri_Gen #(.i(37)) TG37 (f,g,h, fx[w-1-37],gx[w-1-37],hx[w-1-37],     );
 Tri_Gen #(.i(38)) TG38 (f,g,h, fx[w-1-38],gx[w-1-38],hx[w-1-38],     );
 Tri_Gen #(.i(39)) TG39 (f,g,h, fx[w-1-39],gx[w-1-39],hx[w-1-39],     );
 Tri_Gen #(.i(40)) TG40 (f,g,h, fx[w-1-40],gx[w-1-40],hx[w-1-40],     );
 Tri_Gen #(.i(41)) TG41 (f,g,h, fx[w-1-41],gx[w-1-41],hx[w-1-41],     );
 Tri_Gen #(.i(42)) TG42 (f,g,h, fx[w-1-42],gx[w-1-42],hx[w-1-42],     );
 Tri_Gen #(.i(43)) TG43 (f,g,h, fx[w-1-43],gx[w-1-43],hx[w-1-43],     );
 Tri_Gen #(.i(44)) TG44 (f,g,h, fx[w-1-44],gx[w-1-44],hx[w-1-44],     );
 Tri_Gen #(.i(45)) TG45 (f,g,h, fx[w-1-45],gx[w-1-45],hx[w-1-45],     );
 Tri_Gen #(.i(46)) TG46 (f,g,h, fx[w-1-46],gx[w-1-46],hx[w-1-46],     );
 Tri_Gen #(.i(47)) TG47 (f,g,h, fx[w-1-47],gx[w-1-47],hx[w-1-47],     );

 Tri_Gen #(.i(48)) TG48 (f,g,h, fx[w-1-48],gx[w-1-48],hx[w-1-48],     );
 Tri_Gen #(.i(49)) TG49 (f,g,h, fx[w-1-49],gx[w-1-49],hx[w-1-49],     );
 Tri_Gen #(.i(50)) TG50 (f,g,h, fx[w-1-50],gx[w-1-50],hx[w-1-50],     );
 Tri_Gen #(.i(51)) TG51 (f,g,h, fx[w-1-51],gx[w-1-51],hx[w-1-51],     );
 Tri_Gen #(.i(52)) TG52 (f,g,h, fx[w-1-52],gx[w-1-52],hx[w-1-52],     );
 Tri_Gen #(.i(53)) TG53 (f,g,h, fx[w-1-53],gx[w-1-53],hx[w-1-53],     );
 Tri_Gen #(.i(54)) TG54 (f,g,h, fx[w-1-54],gx[w-1-54],hx[w-1-54],     );
 Tri_Gen #(.i(55)) TG55 (f,g,h, fx[w-1-55],gx[w-1-55],hx[w-1-55],     );
 Tri_Gen #(.i(56)) TG56 (f,g,h, fx[w-1-56],gx[w-1-56],hx[w-1-56],     );
 Tri_Gen #(.i(57)) TG57 (f,g,h, fx[w-1-57],gx[w-1-57],hx[w-1-57],     );
 Tri_Gen #(.i(58)) TG58 (f,g,h, fx[w-1-58],gx[w-1-58],hx[w-1-58],     );
 Tri_Gen #(.i(59)) TG59 (f,g,h, fx[w-1-59],gx[w-1-59],hx[w-1-59],     );
 Tri_Gen #(.i(60)) TG60 (f,g,h, fx[w-1-60],gx[w-1-60],hx[w-1-60],     );
 Tri_Gen #(.i(61)) TG61 (f,g,h, fx[w-1-61],gx[w-1-61],hx[w-1-61],     );
 Tri_Gen #(.i(62)) TG62 (f,g,h, fx[w-1-62],gx[w-1-62],hx[w-1-62],     );
 Tri_Gen #(.i(63)) TG63 (f,g,h, fx[w-1-63],gx[w-1-63],hx[w-1-63],     );


 TriviumP_Reg #(.w(w)) REG0 ({f[93-1-w:0],fx},{g[84-1-w:0],gx},{h[111-1-w:0],hx},rekey,pause, fReg,gReg,hReg, clk,aclr);

 TriviumP_Control #(.w(w)) CTRL0 (rekey,pause,{(w){1'b0}},  ,valid, clk,aclr);


// =========================================

endmodule




// *****************************************
//   Trivium - Pipeline Output Stage
//
// *****************************************

module TriviumP_Out (f,g,h,rekey,pause, keystream,valid, clk,aclr);
parameter w = 1;

// =========================================

 input [93-1:0]  f;
 input [84-1:0]  g;
 input [111-1:0] h;
 input  rekey,pause;

 output [w-1:0] keystream;
 output valid;

 input  clk,aclr;

 wire [w-1:0] z;


// =========================================

 Tri_Gen #(.i(0))  TG00 (f,g,h,         ,           ,          , z[0] );
 Tri_Gen #(.i(1))  TG01 (f,g,h,         ,           ,          , z[1] );
 Tri_Gen #(.i(2))  TG02 (f,g,h,         ,           ,          , z[2] );
 Tri_Gen #(.i(3))  TG03 (f,g,h,         ,           ,          , z[3] );
 Tri_Gen #(.i(4))  TG04 (f,g,h,         ,           ,          , z[4] );
 Tri_Gen #(.i(5))  TG05 (f,g,h,         ,           ,          , z[5] );
 Tri_Gen #(.i(6))  TG06 (f,g,h,         ,           ,          , z[6] );
 Tri_Gen #(.i(7))  TG07 (f,g,h,         ,           ,          , z[7] );
 Tri_Gen #(.i(8))  TG08 (f,g,h,         ,           ,          , z[8] );
 Tri_Gen #(.i(9))  TG09 (f,g,h,         ,           ,          , z[9] );
 Tri_Gen #(.i(10)) TG10 (f,g,h,         ,           ,          , z[10]);
 Tri_Gen #(.i(11)) TG11 (f,g,h,         ,           ,          , z[11]);
 Tri_Gen #(.i(12)) TG12 (f,g,h,         ,           ,          , z[12]);
 Tri_Gen #(.i(13)) TG13 (f,g,h,         ,           ,          , z[13]);
 Tri_Gen #(.i(14)) TG14 (f,g,h,         ,           ,          , z[14]);
 Tri_Gen #(.i(15)) TG15 (f,g,h,         ,           ,          , z[15]);

 Tri_Gen #(.i(16)) TG16 (f,g,h,         ,           ,          , z[16]);
 Tri_Gen #(.i(17)) TG17 (f,g,h,         ,           ,          , z[17]);
 Tri_Gen #(.i(18)) TG18 (f,g,h,         ,           ,          , z[18]);
 Tri_Gen #(.i(19)) TG19 (f,g,h,         ,           ,          , z[19]);
 Tri_Gen #(.i(20)) TG20 (f,g,h,         ,           ,          , z[20]);
 Tri_Gen #(.i(21)) TG21 (f,g,h,         ,           ,          , z[21]);
 Tri_Gen #(.i(22)) TG22 (f,g,h,         ,           ,          , z[22]);
 Tri_Gen #(.i(23)) TG23 (f,g,h,         ,           ,          , z[23]);
 Tri_Gen #(.i(24)) TG24 (f,g,h,         ,           ,          , z[24]);
 Tri_Gen #(.i(25)) TG25 (f,g,h,         ,           ,          , z[25]);
 Tri_Gen #(.i(26)) TG26 (f,g,h,         ,           ,          , z[26]);
 Tri_Gen #(.i(27)) TG27 (f,g,h,         ,           ,          , z[27]);
 Tri_Gen #(.i(28)) TG28 (f,g,h,         ,           ,          , z[28]);
 Tri_Gen #(.i(29)) TG29 (f,g,h,         ,           ,          , z[29]);
 Tri_Gen #(.i(30)) TG30 (f,g,h,         ,           ,          , z[30]);
 Tri_Gen #(.i(31)) TG31 (f,g,h,         ,           ,          , z[31]);

 Tri_Gen #(.i(32)) TG32 (f,g,h,         ,           ,          , z[32]);
 Tri_Gen #(.i(33)) TG33 (f,g,h,         ,           ,          , z[33]);
 Tri_Gen #(.i(34)) TG34 (f,g,h,         ,           ,          , z[34]);
 Tri_Gen #(.i(35)) TG35 (f,g,h,         ,           ,          , z[35]);
 Tri_Gen #(.i(36)) TG36 (f,g,h,         ,           ,          , z[36]);
 Tri_Gen #(.i(37)) TG37 (f,g,h,         ,           ,          , z[37]);
 Tri_Gen #(.i(38)) TG38 (f,g,h,         ,           ,          , z[38]);
 Tri_Gen #(.i(39)) TG39 (f,g,h,         ,           ,          , z[39]);
 Tri_Gen #(.i(40)) TG40 (f,g,h,         ,           ,          , z[40]);
 Tri_Gen #(.i(41)) TG41 (f,g,h,         ,           ,          , z[41]);
 Tri_Gen #(.i(42)) TG42 (f,g,h,         ,           ,          , z[42]);
 Tri_Gen #(.i(43)) TG43 (f,g,h,         ,           ,          , z[43]);
 Tri_Gen #(.i(44)) TG44 (f,g,h,         ,           ,          , z[44]);
 Tri_Gen #(.i(45)) TG45 (f,g,h,         ,           ,          , z[45]);
 Tri_Gen #(.i(46)) TG46 (f,g,h,         ,           ,          , z[46]);
 Tri_Gen #(.i(47)) TG47 (f,g,h,         ,           ,          , z[47]);

 Tri_Gen #(.i(48)) TG48 (f,g,h,         ,           ,          , z[48]);
 Tri_Gen #(.i(49)) TG49 (f,g,h,         ,           ,          , z[49]);
 Tri_Gen #(.i(50)) TG50 (f,g,h,         ,           ,          , z[50]);
 Tri_Gen #(.i(51)) TG51 (f,g,h,         ,           ,          , z[51]);
 Tri_Gen #(.i(52)) TG52 (f,g,h,         ,           ,          , z[52]);
 Tri_Gen #(.i(53)) TG53 (f,g,h,         ,           ,          , z[53]);
 Tri_Gen #(.i(54)) TG54 (f,g,h,         ,           ,          , z[54]);
 Tri_Gen #(.i(55)) TG55 (f,g,h,         ,           ,          , z[55]);
 Tri_Gen #(.i(56)) TG56 (f,g,h,         ,           ,          , z[56]);
 Tri_Gen #(.i(57)) TG57 (f,g,h,         ,           ,          , z[57]);
 Tri_Gen #(.i(58)) TG58 (f,g,h,         ,           ,          , z[58]);
 Tri_Gen #(.i(59)) TG59 (f,g,h,         ,           ,          , z[59]);
 Tri_Gen #(.i(60)) TG60 (f,g,h,         ,           ,          , z[60]);
 Tri_Gen #(.i(61)) TG61 (f,g,h,         ,           ,          , z[61]);
 Tri_Gen #(.i(62)) TG62 (f,g,h,         ,           ,          , z[62]);
 Tri_Gen #(.i(63)) TG63 (f,g,h,         ,           ,          , z[63]);


 TriviumP_Control #(.w(w)) CTRL0 (rekey,pause,z, keystream,valid, clk,aclr);


// =========================================

endmodule




// ------------------------------------------------------
// ------------------------------------------------------





// *****************************************
//   Trivium Pipelined - Register Setup
//
// *****************************************

module TriviumP_Reg (load_f,load_g,load_h,rekey,pause, f,g,h, clk,aclr);
parameter w = 1;

// =========================================

 input [93-1:0]  load_f;
 input [84-1:0]  load_g;
 input [111-1:0] load_h;
 input  rekey,pause;

 output reg [93-1:0]  f;
 output reg [84-1:0]  g;
 output reg [111-1:0] h;

 input  clk,aclr;


// =========================================

 always @(posedge clk, posedge aclr) begin
   if(aclr)        f <= 93'b0;
   else if(!rekey) f <= 93'b0;
   else if(!pause) f <= load_f;
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        g <= 84'b0;
   else if(!rekey) g <= 84'b0;
   else if(!pause) g <= load_g;
  end

 always @(posedge clk, posedge aclr) begin
   if(aclr)        h <= 111'b0;
   else if(!rekey) h <= 111'b0;
   else if(!pause) h <= load_h;
  end


// =========================================

endmodule




// *****************************************
//   Trivium Pipelined - Control Unit
//
// *****************************************

module TriviumP_Control (set,pause,z, keystream,valid, clk,aclr);
parameter w = 1;

// =========================================

 input  set,pause;

 output reg valid;

 input  clk,aclr;

 input [w-1:0] z;
 output reg [w-1:0] keystream;


// =========================================

 always@(posedge clk, posedge aclr) begin
   if(aclr)        valid <= 1'b0;
   else if(!set)   valid <= 1'b0;
   else if(!pause) valid <= 1'b1;
  end

 always@(posedge clk, posedge aclr) begin
   if(aclr)         keystream <= {(w){1'b0}};
   else if(!set)    keystream <= {(w){1'b0}};
   else if(!pause)  keystream <= z;
  end


// =========================================

endmodule





// ------------------------------------------------------
// ------------------------------------------------------
