

// *****************************************
//   Stream Cipher
//   Xbit Key, 32bit IV
//
// *****************************************
//   Verilog Code (Oct 07)
//   Iain Devlin 2007
//
// *****************************************

`define SC_IN_WIDTH 16     //max 31
`define SC_NUMBER 2
`define SC_WIDTH 16
`define KEY_WIDTH 128
`define STREAM_CIPHER Grain128X_n2r

// Trivium_1  //fully tested
// Trivium_2  //fully tested
// Trivium_8  //fully tested
// Trivium_16 //fully tested
// Trivium_64 //fully tested
// TriviumP64 //fully tested - (keysearch)
// key0 iv0 = 1101 1111 0000 0111 1111 1101 ...

// Grain_1    //fully tested but algorithm insecure
// Grain_8    //fully tested but algorithm insecure
// Grain_16   //fully tested but algorithm insecure
// GrainP16   //fully tested - (keysearch)
// key0 iv0 = 0100 1100 0101 0000 1010 ...

// Grain1_1   //fully tested
// Grain1_8   //fully tested 
// Grain1_16  //fully tested 
// Grain1p_16 //fully tested - (keysearch)
// key0 iv0 = 0111 1011 1001 0111 1000 ...

// Grain1X_n2 //fully tested - (keysearch)
// Grain1X_n4 //fully tested - (keysearch)
// key[79:78]00 [77:0]0 iv0 = 0111 1011 1001 0111 ...
// key[79:78]10 [77:0]0 iv0 = 0100 0001 0111 1010 ...
// key[79:78]01 [77:0]0 iv0 = 1011 0000 1001 1110 ...
// key[79:78]11 [77:0]0 iv0 = 0100 0011 1111 0001 ...

// Grain128_1    //fully tested
// Grain128_8    //fully tested 
// Grain128_16   //fully tested 
// Grain128_32   //fully tested 
// Grain128_32r  //fully tested 
// Grain128p_16  //fully tested - (keysearch)
// Grain128p_32r //fully tested - (keysearch)
// key0 iv0 = 0000 1111 1101 1001 1101 ...

// Grain128X_n2r //fully tested - (keysearch)
// key[127]0 [126:0]0 iv0 = 0000 1111 1101 1001 1101 ...
// key[127]1 [126:0]0 iv0 = 1110 1000 1100 1010 1111 ...




// *****************************************
//   Stream Cipher Wrapper Module
//
// *****************************************

module Stream_Core (data,addr,wr,rekey,pause, keystream,valid, clk,aclr);
parameter w = `SC_WIDTH, u = `SC_IN_WIDTH, n = `SC_NUMBER, k = `KEY_WIDTH;

// =========================================

 input  [u-1:0] data;
 input  addr;
 input  wr;
 input  rekey;
 input  pause;

 output [w*n-1:0] keystream;
 output [n-1:0] valid;

 input  clk,aclr;

 wire  [k-1:0]  Key;
 wire  [32-1:0]  IV;


// =========================================

 SC_Interface  #(.u(u), .k(k)) SCI (data,addr,wr, Key,IV, clk,aclr);
 `STREAM_CIPHER  #(.w(w)) SC0 (Key,IV,rekey,pause, keystream,valid, clk,aclr);


// =========================================

endmodule





// *****************************************
//   Stream Cipher / Stream Core Interface
//
// *****************************************

module  SC_Interface (data,sel,wr, Key,IV, clk,aclr);
parameter u = 1, k = 80;

// =========================================

 input  [u-1:0] data;
 input  sel;
 input  wr;

 output reg [k-1:0]  Key;
 output reg [32-1:0]  IV;

 input  clk,aclr;


// =========================================

 always@(posedge clk, posedge aclr) begin
   if(aclr) begin
     Key <= {(k){1'b0}};
     IV  <= 32'b0;
    end
   else if(wr) begin
     if(sel) Key <= {Key[k-1-u:0],data};
     else    IV  <= {IV[32-1-u:0],data};
    end
 end


// =========================================

endmodule